from math import cos, pi, sin

from pygfunction.media import Fluid

from ghedesigner.enums import FluidType


class GHEFluid(Fluid):

    def __init__(self, fluid_str: str, percent: float, temperature: float = 20):

        fluid_str = fluid_str.upper()
        if fluid_str == FluidType.ETHYLALCOHOL.name:
            self.fluid_type = FluidType.ETHYLALCOHOL
        elif fluid_str == FluidType.ETHYLENEGLYCOL.name:
            self.fluid_type = FluidType.ETHYLENEGLYCOL
        elif fluid_str == FluidType.METHYLALCOHOL.name:
            self.fluid_type = FluidType.METHYLALCOHOL
        elif fluid_str == FluidType.PROPYLENEGLYCOL.name:
            self.fluid_type = FluidType.PROPYLENEGLYCOL
        elif fluid_str == FluidType.WATER.name:
            self.fluid_type = FluidType.WATER
        else:
            raise ValueError(f"FluidType \"{fluid_str}\" not implemented")

        fluid_map = {
            FluidType.ETHYLALCOHOL.name: "MEA",
            FluidType.ETHYLENEGLYCOL.name: "MEG",
            FluidType.METHYLALCOHOL.name: "MMA",
            FluidType.PROPYLENEGLYCOL.name: "MPG",
            FluidType.WATER.name: "WATER"
        }

        super().__init__(fluid_map[fluid_str], percent, temperature)
        self.concentration_percent = percent
        self.temperature = temperature

    def to_input(self):
        return {'fluid_name': self.fluid_type.name,
                'concentration_percent': self.concentration_percent,
                'temperature': self.temperature}


class ThermalProperty:
    def __init__(self, k, rho_cp):
        self.k = k  # Thermal conductivity (W/m.K)
        self.rhoCp = rho_cp  # Volumetric heat capacity (J/K.m3)

    def as_dict(self) -> dict:
        output = dict()
        output['type'] = str(self.__class__)
        output['thermal_conductivity'] = {'value': self.k, 'units': 'W/m-K'}
        output['volumetric_heat_capacity'] = {'value': self.rhoCp, 'units': 'J/K-m3'}
        return output

    def to_input(self) -> dict:
        return {'conductivity': self.k, 'rho_cp': self.rhoCp}


class Grout(ThermalProperty):
    pass


class Pipe(ThermalProperty):
    def __init__(self, pos, r_in, r_out, s, roughness, k, rho_cp):
        # Make variables from ThermalProperty available to Pipe
        super().__init__(k, rho_cp)

        # Pipe specific parameters
        self.pos = pos  # Pipe positions either a list of tuples or tuple
        self.r_in = r_in  # Pipe inner radius (m) can be a float or list
        self.r_out = r_out  # Pipe outer radius (m) can be a float or list
        self.s = s  # Center pipe to center pipe shank spacing
        self.roughness = roughness  # Pipe roughness (m)
        if type(pos) is list:
            self.n_pipes = int(len(pos) / 2)  # Number of pipes
        else:
            self.n_pipes = 1

    def as_dict(self) -> dict:
        output = dict()
        output['base'] = super().as_dict()
        output['pipe_center_positions'] = str(self.pos)
        if type(self.r_in) is float:
            output["pipe_inner_diameter"] = str(self.r_in * 2.0)
            output["pipe_outer_diameter"] = str(self.r_out * 2.0)
        else:
            output["pipe_inner_diameters"] = str([x * 2.0 for x in self.r_in])
            output["pipe_outer_diameters"] = str([x * 2.0 for x in self.r_out])
        output['shank_spacing_pipe_to_pipe'] = {'value': self.s, 'units': 'm'}
        output['pipe_roughness'] = {'value': self.roughness, 'units': 'm'}
        output['number_of_pipes'] = self.n_pipes
        return output

    @staticmethod
    def place_pipes(s, r_out, n_pipes):
        """Positions pipes in an axis-symmetric configuration."""
        shank_space = s / 2 + r_out
        dt = pi / float(n_pipes)
        pos = [(0.0, 0.0) for _ in range(2 * n_pipes)]
        for i in range(n_pipes):
            pos[2 * i] = (shank_space * cos(2.0 * i * dt + pi), shank_space * sin(2.0 * i * dt + pi))
            pos[2 * i + 1] = (shank_space * cos(2.0 * i * dt + pi + dt), shank_space * sin(2.0 * i * dt + pi + dt))
        return pos


class Soil(ThermalProperty):
    def __init__(self, k, rho_cp, ugt):
        # Make variables from ThermalProperty available to Pipe
        ThermalProperty.__init__(self, k, rho_cp)

        # Soil specific parameters
        self.ugt = ugt

    def as_dict(self) -> dict:
        output = super().as_dict()
        output['undisturbed_ground_temperature'] = {'value': self.ugt, 'units': 'C'}
        return output

    def to_input(self) -> dict:
        return {'conductivity': self.k, 'rho_cp': self.rhoCp, 'undisturbed_temp': self.ugt}



from abc import abstractmethod
from math import floor
from typing import Union

from ghedesigner.borehole import GHEBorehole
from ghedesigner.domains import polygonal_land_constraint, bi_rectangle_nested
from ghedesigner.domains import square_and_near_square, rectangular, bi_rectangle_zoned_nested
from ghedesigner.enums import BHPipeType, TimestepType, FlowConfigType
from ghedesigner.geometry import GeometricConstraints, GeometricConstraintsBiRectangle
from ghedesigner.geometry import GeometricConstraintsBiZoned, GeometricConstraintsBiRectangleConstrained
from ghedesigner.geometry import GeometricConstraintsNearSquare, GeometricConstraintsRectangle
from ghedesigner.geometry import GeometricConstraintsRowWise
from ghedesigner.media import Grout, Pipe, Soil, GHEFluid
from ghedesigner.search_routines import Bisection1D, Bisection2D, BisectionZD, RowWiseModifiedBisectionSearch
from ghedesigner.simulation import SimulationParameters

AnyBisectionType = Union[Bisection1D, Bisection2D, BisectionZD, RowWiseModifiedBisectionSearch]


class DesignBase:
    def __init__(
            self,
            v_flow: float,
            _borehole: GHEBorehole,
            bhe_type: BHPipeType,
            fluid: GHEFluid,
            pipe: Pipe,
            grout: Grout,
            soil: Soil,
            sim_params: SimulationParameters,
            geometric_constraints: GeometricConstraints,
            hourly_extraction_ground_loads: list,
            method: TimestepType,
            flow_type: FlowConfigType = FlowConfigType.BOREHOLE,
            load_years=None
    ):
        if load_years is None:
            load_years = [2019]
        self.load_years = load_years
        self.V_flow = v_flow  # volumetric flow rate, m3/s
        self.borehole = _borehole
        self.bhe_type = bhe_type  # a borehole heat exchanger object
        self.fluid = fluid  # a fluid object
        self.pipe = pipe
        self.grout = grout
        self.soil = soil
        self.sim_params = sim_params
        self.geometric_constraints = geometric_constraints
        self.hourly_extraction_ground_loads = hourly_extraction_ground_loads
        self.method = method
        self.flow_type = flow_type
        if self.method == "hourly":
            msg = (
                "Note: It is not recommended to perform a field selection \n",
                "with the hourly simulation due to computation time. If \n",
                "the goal is to validate the selected field with the \n",
                "hourly simulation, the better solution is to utilize the \n",
                "hybrid simulation to automatically select the field. Then \n",
                "perform a sizing routine on the selected GHE with the \n",
                "hourly simulation."
            )
            # Wrap the text to a 50 char line width and print it
            for element in msg:
                print(element)
            print("\n")

    @abstractmethod
    def find_design(self, disp=False) -> AnyBisectionType:
        pass

    def to_input(self) -> dict:
        return {'flow_rate': self.V_flow, 'flow_type': self.flow_type.name}


class DesignNearSquare(DesignBase):
    def __init__(self, v_flow: float, _borehole: GHEBorehole, bhe_type: BHPipeType,
                 fluid: GHEFluid, pipe: Pipe, grout: Grout, soil: Soil, sim_params: SimulationParameters,
                 geometric_constraints: GeometricConstraintsNearSquare, hourly_extraction_ground_loads: list,
                 method: TimestepType, flow_type: FlowConfigType = FlowConfigType.BOREHOLE, load_years=None):
        super().__init__(v_flow, _borehole, bhe_type, fluid, pipe, grout, soil, sim_params, geometric_constraints,
                         hourly_extraction_ground_loads, method, flow_type, load_years)
        self.geometric_constraints = geometric_constraints
        # If a near-square design routine is requested, then we go from a
        # 1x1 to 32x32 at the B-spacing
        # The lower end of the near-square routine is always 1
        # There would never be a time that a user would __need__ to give a
        # different lower range. The upper number of boreholes range is
        # calculated based on the spacing and length provided.
        n = floor(self.geometric_constraints.length / self.geometric_constraints.b) + 1
        number_of_boreholes = int(n)
        self.coordinates_domain, self.fieldDescriptors = square_and_near_square(1, number_of_boreholes,
                                                                                self.geometric_constraints.b)

    def find_design(self, disp=False) -> Bisection1D:
        if disp:
            title = "Find near-square.."
            print(title + "\n" + len(title) * "=")
        return Bisection1D(
            self.coordinates_domain,
            self.fieldDescriptors,
            self.V_flow,
            self.borehole,
            self.bhe_type,
            self.fluid,
            self.pipe,
            self.grout,
            self.soil,
            self.sim_params,
            self.hourly_extraction_ground_loads,
            method=self.method,
            flow_type=self.flow_type,
            disp=disp,
            field_type="near-square",
            load_years=self.load_years,
        )


class DesignRectangle(DesignBase):
    def __init__(self, v_flow: float, _borehole: GHEBorehole, bhe_type: BHPipeType,
                 fluid: GHEFluid, pipe: Pipe, grout: Grout, soil: Soil, sim_params: SimulationParameters,
                 geometric_constraints: GeometricConstraintsRectangle, hourly_extraction_ground_loads: list,
                 method: TimestepType, flow_type: FlowConfigType = FlowConfigType.BOREHOLE, load_years=None):
        super().__init__(v_flow, _borehole, bhe_type, fluid, pipe, grout, soil, sim_params, geometric_constraints,
                         hourly_extraction_ground_loads, method, flow_type, load_years)
        self.geometric_constraints = geometric_constraints
        self.coordinates_domain, self.fieldDescriptors = rectangular(
            self.geometric_constraints.length, self.geometric_constraints.width,
            self.geometric_constraints.b_min, self.geometric_constraints.b_max_x)

    def find_design(self, disp=False) -> Bisection1D:
        if disp:
            title = "Find rectangle..."
            print(title + "\n" + len(title) * "=")
        return Bisection1D(
            self.coordinates_domain,
            self.fieldDescriptors,
            self.V_flow,
            self.borehole,
            self.bhe_type,
            self.fluid,
            self.pipe,
            self.grout,
            self.soil,
            self.sim_params,
            self.hourly_extraction_ground_loads,
            method=self.method,
            flow_type=self.flow_type,
            disp=disp,
            field_type="rectangle",
            load_years=self.load_years,
        )


class DesignBiRectangle(DesignBase):
    def __init__(self, v_flow: float, _borehole: GHEBorehole, bhe_type: BHPipeType,
                 fluid: GHEFluid, pipe: Pipe, grout: Grout, soil: Soil, sim_params: SimulationParameters,
                 geometric_constraints: GeometricConstraintsBiRectangle, hourly_extraction_ground_loads: list,
                 method: TimestepType, flow_type: FlowConfigType = FlowConfigType.BOREHOLE, load_years=None):
        super().__init__(v_flow, _borehole, bhe_type, fluid, pipe, grout, soil, sim_params, geometric_constraints,
                         hourly_extraction_ground_loads, method, flow_type, load_years)
        self.geometric_constraints = geometric_constraints
        self.coordinates_domain_nested, self.fieldDescriptors = bi_rectangle_nested(
            self.geometric_constraints.length, self.geometric_constraints.width, self.geometric_constraints.b_min,
            self.geometric_constraints.b_max_x, self.geometric_constraints.b_max_y, disp=False
        )

    def find_design(self, disp=False) -> Bisection2D:
        if disp:
            title = "Find bi-rectangle..."
            print(title + "\n" + len(title) * "=")
        return Bisection2D(
            self.coordinates_domain_nested,
            self.fieldDescriptors,
            self.V_flow,
            self.borehole,
            self.bhe_type,
            self.fluid,
            self.pipe,
            self.grout,
            self.soil,
            self.sim_params,
            self.hourly_extraction_ground_loads,
            method=self.method,
            flow_type=self.flow_type,
            disp=disp,
            field_type="bi-rectangle",
            load_years=self.load_years,
        )


class DesignBiZoned(DesignBase):
    def __init__(self, v_flow: float, _borehole: GHEBorehole, bhe_type: BHPipeType,
                 fluid: GHEFluid, pipe: Pipe, grout: Grout, soil: Soil, sim_params: SimulationParameters,
                 geometric_constraints: GeometricConstraintsBiZoned, hourly_extraction_ground_loads: list,
                 method: TimestepType, flow_type: FlowConfigType = FlowConfigType.BOREHOLE, load_years=None):
        super().__init__(v_flow, _borehole, bhe_type, fluid, pipe, grout, soil, sim_params, geometric_constraints,
                         hourly_extraction_ground_loads, method, flow_type, load_years)
        self.geometric_constraints = geometric_constraints
        self.coordinates_domain_nested, self.fieldDescriptors = bi_rectangle_zoned_nested(
            self.geometric_constraints.length, self.geometric_constraints.width, self.geometric_constraints.b_min,
            self.geometric_constraints.b_max_x, self.geometric_constraints.b_max_y
        )

    def find_design(self, disp=False) -> BisectionZD:
        if disp:
            title = "Find bi-zoned..."
            print(title + "\n" + len(title) * "=")
        return BisectionZD(
            self.coordinates_domain_nested,
            self.fieldDescriptors,
            self.V_flow,
            self.borehole,
            self.bhe_type,
            self.fluid,
            self.pipe,
            self.grout,
            self.soil,
            self.sim_params,
            self.hourly_extraction_ground_loads,
            method=self.method,
            flow_type=self.flow_type,
            disp=disp,
            field_type="bi-zoned",
        )


class DesignBiRectangleConstrained(DesignBase):
    def __init__(self, v_flow: float, _borehole: GHEBorehole, bhe_type: BHPipeType,
                 fluid: GHEFluid, pipe: Pipe, grout: Grout, soil: Soil, sim_params: SimulationParameters,
                 geometric_constraints: GeometricConstraintsBiRectangleConstrained,
                 hourly_extraction_ground_loads: list, method: TimestepType,
                 flow_type: FlowConfigType = FlowConfigType.BOREHOLE, load_years=None):
        super().__init__(v_flow, _borehole, bhe_type, fluid, pipe, grout, soil, sim_params, geometric_constraints,
                         hourly_extraction_ground_loads, method, flow_type, load_years)
        self.geometric_constraints = geometric_constraints
        self.coordinates_domain_nested, self.fieldDescriptors = polygonal_land_constraint(
            self.geometric_constraints.b_min,
            self.geometric_constraints.b_max_x,
            self.geometric_constraints.b_max_y,
            self.geometric_constraints.property_boundary,
            self.geometric_constraints.no_go_boundaries,
        )

    def find_design(self, disp=False) -> Bisection2D:
        if disp:
            title = "Find bi-rectangle_constrained..."
            print(title + "\n" + len(title) * "=")
        return Bisection2D(
            self.coordinates_domain_nested,
            self.fieldDescriptors,
            self.V_flow,
            self.borehole,
            self.bhe_type,
            self.fluid,
            self.pipe,
            self.grout,
            self.soil,
            self.sim_params,
            self.hourly_extraction_ground_loads,
            method=self.method,
            flow_type=self.flow_type,
            disp=disp,
            field_type="bi-rectangle_constrained",
            load_years=self.load_years,
        )


class DesignRowWise(DesignBase):
    def __init__(self, v_flow: float, _borehole: GHEBorehole, bhe_type: BHPipeType,
                 fluid: GHEFluid, pipe: Pipe, grout: Grout, soil: Soil, sim_params: SimulationParameters,
                 geometric_constraints: GeometricConstraintsRowWise, hourly_extraction_ground_loads: list,
                 method: TimestepType, flow_type: FlowConfigType = FlowConfigType.BOREHOLE, load_years=None):
        super().__init__(v_flow, _borehole, bhe_type, fluid, pipe, grout, soil, sim_params, geometric_constraints,
                         hourly_extraction_ground_loads, method, flow_type, load_years)
        self.geometric_constraints = geometric_constraints

    def find_design(self, disp=False) -> RowWiseModifiedBisectionSearch:
        if disp:
            title = "Find row-wise..."
            print(title + "\n" + len(title) * "=")
        return RowWiseModifiedBisectionSearch(
            self.V_flow,
            self.borehole,
            self.bhe_type,
            self.fluid,
            self.pipe,
            self.grout,
            self.soil,
            self.sim_params,
            self.hourly_extraction_ground_loads,
            self.geometric_constraints,
            method=self.method,
            flow_type=self.flow_type,
            disp=disp,
            field_type="row-wise",
            load_years=self.load_years,
        )



from math import ceil, floor

from ghedesigner.coordinates import rectangle, transpose_coordinates, zoned_rectangle, l_shape, c_shape, lop_u
from ghedesigner.feature_recognition import remove_cutout, determine_largest_rectangle


def square_and_near_square(lower: int, upper: int, b: float):
    if lower < 1 or upper < 1:
        raise ValueError("The lower and upper arguments must be positive" "integer values.")
    if upper < lower:
        raise ValueError("The lower argument should be less than or equal to" "the upper.")

    field_descriptors = []
    coordinates_domain = []
    # field_descriptors = ["1X1", "1X2", "1X3"]
    # coordinates_domain = [
    #     [[0, 0]],
    #     [[0, 0], [0, b]],
    #     [[0, 0], [0, b], [0, 2 * b]]
    # ]

    for i in range(lower, upper + 1):
        for j in range(2):
            coordinates = rectangle(i, i + j, b, b)

            coordinates_domain.append(coordinates)
            field_descriptors.append(f"{i}X{i + j}")

    return coordinates_domain, field_descriptors


def rectangular(length_x: float, length_y: float, b_min: float, b_max: float, disp: bool = False):
    # Make this work for the transpose
    if length_x >= length_y:
        length_1 = length_x
        length_2 = length_y
        transpose = False
    else:
        length_1 = length_y
        length_2 = length_x
        transpose = True

    rectangle_domain = []
    field_descriptors = []
    # find the maximum number of boreholes as a float
    n_1_max = (length_1 / b_min) + 1
    n_1_min = (length_1 / b_max) + 1

    n_min = ceil(n_1_min)
    n_max = floor(n_1_max)

    n_2_old = 1

    if disp:
        print(50 * "-")
        print("Rectangular Domain\nNx\tNy\tBx\tBy")

    _iter = 0

    for N in range(n_min, n_max + 1):
        # Check to see if we bracket
        b = length_1 / (N - 1)
        n_2 = floor((length_2 / b) + 1)

        if _iter == 0:
            for i in range(1, n_min):
                r = rectangle(i, 1, b, b)
                if transpose:
                    r = transpose_coordinates(r)
                rectangle_domain.append(r)
                field_descriptors.append(f"{i}X{1}_B{b:0.2f}")
            for j in range(1, n_2):
                r = rectangle(n_min, j, b, b)
                if transpose:
                    r = transpose_coordinates(r)
                rectangle_domain.append(r)
                field_descriptors.append(f"{n_min}X{j}_B{b:0.2f}")

            _iter += 1
        if n_2_old == n_2:
            pass
        else:
            r = rectangle(N, n_2, b, b)
            if disp:
                print(f"{N}\t{n_2}\t{b}\t{b}")
            if transpose:
                r = transpose_coordinates(r)
            rectangle_domain.append(r)
            field_descriptors.append(f"{N}X{n_2}_B{b:0.2f}")
            n_2_old = n_2

        N += 1

    return rectangle_domain, field_descriptors


def bi_rectangular(length_x, length_y, b_min, b_max_x, b_max_y, transpose=False, disp=False):
    # Make this work for the transpose
    if length_x >= length_y:
        length_1 = length_x
        length_2 = length_y
        b_max_1 = b_max_x
        b_max_2 = b_max_y
    else:
        length_1 = length_y
        length_2 = length_x
        b_max_1 = b_max_y
        b_max_2 = b_max_x

    bi_rectangle_domain = []
    field_descriptors = []
    # find the maximum number of boreholes as a float
    n_1_max = (length_1 / b_min) + 1
    n_1_min = (length_1 / b_max_1) + 1

    # if it is the first case in the domain, we want to step up from one
    # borehole, to a line, to adding the rows
    _iter = 0

    n_min = ceil(n_1_min)
    n_max = floor(n_1_max)
    for n_1 in range(n_min, n_max + 1):

        n_2 = ceil((length_2 / b_max_2) + 1)
        b_2 = length_2 / (n_2 - 1)

        b_1 = length_1 / (n_1 - 1)

        if _iter == 0:
            for i in range(1, n_1):
                coordinates = rectangle(i, 1, b_1, b_2)
                if transpose:
                    coordinates = transpose_coordinates(coordinates)
                bi_rectangle_domain.append(coordinates)
                field_descriptors.append(f"{i}X{1}_B1{b_1:0.2f}_B2{b_2:0.2f}")
            for j in range(1, n_2):
                coordinates = rectangle(n_1, j, b_1, b_2)
                if transpose:
                    coordinates = transpose_coordinates(coordinates)
                bi_rectangle_domain.append(coordinates)
                field_descriptors.append(f"{n_1}X{j}_B1{b_1:0.2f}_B2{b_2:0.2f}")

            _iter += 1

        if disp:
            print(f"{n_1}x{n_2} with {b_1:0.1f}x{b_2:0.1f}")

        coordinates = rectangle(n_1, n_2, b_1, b_2)
        if transpose:
            coordinates = transpose_coordinates(coordinates)
        bi_rectangle_domain.append(coordinates)
        field_descriptors.append(f"{n_1}X{n_2}_B1{b_1:0.2f}_B2{b_2:0.2f}")

        n_1 += 1

    return bi_rectangle_domain, field_descriptors


def bi_rectangle_nested(length_x, length_y, b_min, b_max_x, b_max_y, disp=False):
    # Make this work for the transpose
    if length_x >= length_y:
        length_1 = length_x
        length_2 = length_y
        b_max_1 = b_max_x
        b_max_2 = b_max_y
        transpose = False
    else:
        length_1 = length_y
        length_2 = length_x
        b_max_1 = b_max_y
        b_max_2 = b_max_x
        transpose = True

    # find the maximum number of boreholes as a float
    n_2_max = (length_2 / b_min) + 1
    n_2_min = (length_2 / b_max_2) + 1

    n_min = ceil(n_2_min)
    n_max = floor(n_2_max)

    bi_rectangle_nested_domain = []
    field_descriptors = []

    for n_2 in range(n_min, n_max + 1):
        b_2 = length_2 / (n_2 - 1)
        bi_rectangle_domain, f_d = bi_rectangular(length_1, length_2, b_min, b_max_1,
                                                  b_2, transpose=transpose, disp=disp)
        # print("Bi-Rectangular: ",bi_rectangle_domain)
        bi_rectangle_nested_domain.append(bi_rectangle_domain)
        # fieldDescriptors.append(
        #     str(length_1) + "X" + str(length_2) + "_" + str(B_min) + "_" + str(B_max_1)+"_"+str(b_2)
        # )
        field_descriptors.append(f_d)

    return bi_rectangle_nested_domain, field_descriptors


def zoned_rectangle_domain(length_x, length_y, n_x, n_y, transpose=False):
    # Make this work for the transpose
    if length_x >= length_y:
        length_1 = length_x
        length_2 = length_y
        n_1 = n_x
        n_2 = n_y
    else:
        length_1 = length_y
        length_2 = length_x
        n_1 = n_y
        n_2 = n_x

    b_1 = length_1 / (n_1 - 1)
    b_2 = length_2 / (n_2 - 1)

    _zoned_rectangle_domain = []
    field_descriptors = []

    n_i1 = 1
    n_i2 = 1

    z = zoned_rectangle(n_1, n_2, b_1, b_2, n_i1, n_i2)
    _zoned_rectangle_domain.append(z)
    field_descriptors.append(f"{n_1}X{n_2}_{n_i1}X{n_i2}_B1{b_1:0.2f}_B2{b_2:0.2f}")

    while n_i1 < (n_1 - 2) or n_i2 < (n_2 - 2):

        ratio = b_1 / b_2

        # general case where we can reduce in either direction
        # inner rectangular spacing
        bi_1 = (n_1 - 1) * b_1 / (n_i1 + 1)
        # bi_2 = (n_2 - 1) * b_2 / (n_i2 + 1)
        # inner spacings for increasing each row
        # bi_1_p1 = (n_1 - 1) * b_1 / (n_i1 + 2)
        bi_2_p1 = (n_2 - 1) * b_2 / (n_i2 + 2)

        ratio_1 = bi_1 / bi_2_p1
        # ratio_2 = bi_2 / bi_1_p1

        # we only want to increase one at a time, and we want to increase
        # the one that will keep the inner rectangle furthest from the perimeter

        if ratio_1 > ratio:
            n_i1 += 1
        elif ratio_1 <= ratio:
            n_i2 += 1
        else:
            raise ValueError(
                "This function should not have ever made it to "
                "this point, there may be a problem with the "
                "inputs."
            )
        z = zoned_rectangle(n_1, n_2, b_1, b_2, n_i1, n_i2)
        if transpose:
            z = transpose_coordinates(z)
        _zoned_rectangle_domain.append(z)
        field_descriptors.append(f"{n_1}X{n_2}_{n_i1}X{n_i2}_B1{b_1:0.2f}_B2{b_2:0.2f}")

    return _zoned_rectangle_domain, field_descriptors


def bi_rectangle_zoned_nested(length_x, length_y, b_min, b_max_x, b_max_y):
    # Make this work for the transpose
    if length_x >= length_y:
        length_1 = length_x
        length_2 = length_y
        b_max_1 = b_max_x
        b_max_2 = b_max_y
        transpose = False
    else:
        length_1 = length_y
        length_2 = length_x
        b_max_1 = b_max_y
        b_max_2 = b_max_x
        transpose = True

    # find the maximum number of boreholes as a float
    n_1_max = (length_1 / b_min) + 1
    n_1_min = (length_1 / b_max_1) + 1

    n_2_max = (length_2 / b_min) + 1
    n_2_min = (length_2 / b_max_2) + 1

    n_min_1 = ceil(n_1_min)
    n_max_1 = floor(n_1_max)

    n_min_2 = ceil(n_2_min)
    n_max_2 = floor(n_2_max)

    bi_rectangle_zoned_nested_domain = []
    field_descriptors = []

    n_1_values = list(range(n_min_1, n_max_1 + 1))
    n_2_values = list(range(n_min_2, n_max_2 + 1))

    j = 0  # pertains to n_1_values
    k = 0  # pertains to n_2_values
    index_l = 0

    domain = []
    f_d = []
    for i in range(len(n_1_values) + len(n_2_values) - 1):
        if index_l == 0:
            b_x = length_x / (n_min_1 - 1)
            b_y = length_y / (n_min_2 - 1)

            # go from one borehole to a line
            for index_l in range(1, n_min_1 + 1):
                r = rectangle(index_l, 1, b_x, b_y)
                if transpose:
                    r = transpose_coordinates(r)
                domain.append(r)
                f_d.append(f"{index_l}X{1}_{b_x:0.2f}X{b_y:0.2f}")

            # go from a line to an L
            for index_l in range(2, n_min_2 + 1):
                l_shape_object = l_shape(n_min_1, index_l, b_x, b_y)
                if transpose:
                    l_shape_object = transpose_coordinates(l_shape_object)
                domain.append(l_shape_object)
                f_d.append(f"{n_min_1}X{index_l}_{b_x:0.2f}X{b_y:0.2f}")

            # go from an L to a U
            for index_l in range(2, n_min_2 + 1):
                lop_u_field = lop_u(n_min_1, n_min_2, b_x, b_y, index_l)
                if transpose:
                    lop_u_field = transpose_coordinates(lop_u_field)
                domain.append(lop_u_field)
                f_d.append(f"{n_min_1}X{n_min_2}_{b_x:0.2f}X{b_y:0.2f}")

            # go from a U to an open
            for index_l in range(1, n_min_1 - 1):
                c = c_shape(n_min_1, n_min_2, b_x, b_y, index_l)
                if transpose:
                    c = transpose_coordinates(c)
                domain.append(c)
                f_d.append(f"{n_min_1}X{n_min_2}_{b_x:0.2f}X{b_y:0.2f}")

            index_l += 1

        if i % 2 == 0:
            bi_rectangle_zoned_domain, f_ds = zoned_rectangle_domain(length_1, length_2, n_1_values[j], n_2_values[k],
                                                                     transpose=transpose)
            domain.extend(bi_rectangle_zoned_domain)
            f_d.extend(f_ds)
            if j < len(n_1_values) - 1:
                j += 1
            else:
                k += 1
        else:
            bi_rectangle_zoned_domain, f_ds = zoned_rectangle_domain(length_1, length_2, n_1_values[j], n_2_values[k],
                                                                     transpose=transpose)
            domain.extend(bi_rectangle_zoned_domain)
            f_d.extend(f_ds)
            if k < len(n_2_values) - 1:
                k += 1
            else:
                j += 1

    bi_rectangle_zoned_nested_domain.append(domain)
    field_descriptors.append(f_d)

    return bi_rectangle_zoned_nested_domain, field_descriptors


def polygonal_land_constraint(b_min, b_max_x, b_max_y, property_boundary, no_go_boundaries=None):
    if no_go_boundaries is None:
        no_go_boundaries = []

    outer_rectangle = determine_largest_rectangle(property_boundary)

    x, y = list(zip(*outer_rectangle))
    length = max(x)
    width = max(y)
    coordinates_domain_nested, field_descriptors = bi_rectangle_nested(length, width, b_min, b_max_x, b_max_y)

    coordinates_domain_nested_cutout = []

    for domain in coordinates_domain_nested:
        new_coordinates_domain = []
        for coordinates in domain:
            # Remove boreholes outside of property
            new_coordinates = remove_cutout(coordinates, property_boundary, remove_inside=False)
            # Remove boreholes inside of building
            if len(new_coordinates) == 0:
                continue
            for no_go_zone in no_go_boundaries:
                new_coordinates = remove_cutout(new_coordinates, no_go_zone, remove_inside=True,
                                                keep_contour=False)
            new_coordinates_domain.append(new_coordinates)
        coordinates_domain_nested_cutout.append(new_coordinates_domain)

    coordinates_domain_nested_cutout_reordered = []
    field_descriptors_reordered = []
    for idx, domain in enumerate(coordinates_domain_nested_cutout):
        domain_reordered, f_d_reordered = reorder_domain(domain, field_descriptors[idx])
        coordinates_domain_nested_cutout_reordered.append(domain_reordered)
        field_descriptors_reordered.append(f_d_reordered)

    return coordinates_domain_nested_cutout_reordered, field_descriptors_reordered


def reorder_domain(domain, descriptors):
    """
    Sort domains by length. Rearrange descriptors accordingly.
    Solution from: https://stackoverflow.com/a/9764364

    # TODO: Investigate whether this is needed.
    # TODO: Domains may already be presorted by the nature of the preceding algorithms.
    """

    return zip(*sorted(zip(domain, descriptors), key=lambda x: len(x[0])))



from math import atan, cos, pi, sin, sqrt

import numpy as np

from ghedesigner.constants import DEG_TO_RAD, RAD_TO_DEG, PI_OVER_2
from ghedesigner.shape import Shapes, sort_intersections


def gen_shape(prop_bound, ng_zones=None):
    """Returns an array of shapes objects representing the coordinates given"""
    r_a = [Shapes(prop_bound)]
    if ng_zones is not None:
        r_n = []
        for ngZone in ng_zones:
            r_n.append(Shapes(ngZone))
        r_a.append(r_n)
    else:
        r_a.append(None)
    return r_a


def field_optimization_wp_space_fr(
        p_space,
        space_start,
        rotate_step,
        prop_bound,
        ng_zones=None,
        rotate_start=None,
        rotate_stop=None,
):
    """Optimizes a Field by iterating over input values w/o perimeter spacing

    Parameters: p_space(float): Ratio of perimeter spacing to other target spacing space_start(float): the initial
    target spacing that the optimization program will start with rotate_step(float): the amount of rotation that will
    be changed per step (in degrees) prop_bound([[float,float]]): 2d array of floats that represent the property
    boundary (counter clockwise) ng_zones([[[float,float]]]): 3d array representing the different zones on the
    property where no boreholes can be placed rotate_start(float): the rotation that the field will start at (-pi/2 <
    rotateStart < pi/2) rotate_stop(float): the rotation that the field will stop at (exclusive) (-pi/2 < rotateStop
    < pi/2)

    Outputs: CSVs containing the coordinates for the max field for each target spacing, their respective graphs,
    and their respective data

    """
    if rotate_start is None:
        rotate_start = (-90.0 + rotate_step) * DEG_TO_RAD
    if rotate_stop is None:
        rotate_stop = PI_OVER_2
    if rotate_start > PI_OVER_2 or rotate_start < -PI_OVER_2 or rotate_stop > PI_OVER_2 or rotate_stop < -PI_OVER_2:
        raise ValueError("Invalid Rotation")

    space = space_start
    rt = rotate_start

    y_s = space
    x_s = y_s

    max_l = 0
    max_hole = None
    max_rt = None

    while rt < rotate_stop:
        hole = two_space_gen_bhc(
            prop_bound,
            y_s,
            x_s,
            rotate=rt,
            no_go=ng_zones,
            p_space=p_space * x_s,
            intersection_tolerance=1e-5,
        )

        # Assuming that the rotation with the maximum number of boreholes is most efficiently using space
        if len(hole) > max_l:
            max_l = len(hole)
            max_rt = rt * RAD_TO_DEG
            max_hole = hole

        rt += rotate_step * DEG_TO_RAD

    # Ensures that there are no repeated boreholes
    max_hole = np.array(remove_duplicates(max_hole, p_space * x_s))

    field = max_hole
    field_name = "P" + str(p_space) + "_S" + str(space) + "_rt" + str(max_rt)
    return [field, field_name]


def field_optimization_fr(
        space_start,
        rotate_step,
        prop_bound,
        ng_zones=None,
        rotate_start=None,
        rotate_stop=None,
        intersection_tolerance=1e-5,
):
    """Optimizes a Field by iterating over input values w/o perimeter spacing

    Parameters: space_start(float): the initial target spacing that the optimization program will start with
    rotate_step(float): the amount of rotation that will be changed per step (in degrees) prop_bound([[float,
    float]]): 2d array of floats that represent the property boundary (counter clockwise) ng_zones([[[float,
    float]]]): 3d array representing the different zones on the property where no boreholes can be placed
    rotate_start(float): the rotation that the field will start at (-pi/2 < rotateStart < pi/2) rotate_stop(float):
    the rotation that the field will stop at (exclusive) (-pi/2 < rotateStop < pi/2) intersection_tolerance:

    Outputs: CSVs containing the coordinates for the max field for each target spacing, their respective graphs,
    and their respective data

    """
    if rotate_start is None:
        rotate_start = -90.0 * DEG_TO_RAD
    if rotate_stop is None:
        rotate_stop = PI_OVER_2
    if (
            rotate_start > PI_OVER_2
            or rotate_start < -PI_OVER_2
            or rotate_stop > PI_OVER_2
            or rotate_stop < -PI_OVER_2
    ):
        raise ValueError("Invalid Rotation")

    # Target Spacing iterates

    space = space_start
    rt = rotate_start

    y_s = space
    x_s = y_s

    max_l = 0
    max_hole = None
    max_rt = None

    while rt < rotate_stop:
        hole = gen_borehole_config(
            prop_bound,
            y_s,
            x_s,
            rotate=rt,
            no_go=ng_zones,
            intersection_tolerance=intersection_tolerance,
        )

        # Assuming that the rotation with the maximum number of boreholes is most efficiently using space
        if len(hole) > max_l:
            max_l = len(hole)
            max_rt = rt * RAD_TO_DEG
            max_hole = hole

        rt += rotate_step * DEG_TO_RAD

    # Ensures that there are no repeated boreholes
    max_hole = np.array(remove_duplicates(max_hole, x_s * 1.2))

    field = max_hole
    field_name = "S" + str(space) + "_rt" + str(max_rt)
    return [field, field_name]


def find_duplicates(borefield, space, disp=False):
    """
    The distance method :func:`Borehole.distance` is utilized to find all
    duplicate boreholes in a boreField.
    This function considers a duplicate to be any pair of points that fall
    within each other's radius. The lower index (i) is always stored in the
    0 position of the tuple, while the higher index (j) is stored in the 1
    position.
    Parameters
    ----------
    borefield : list
        A list of :class:`Borehole` objects
    space:
    disp : bool, optional
        Set to true to print progression messages.
        Default is False.
    Returns
    -------
    duplicate_pairs : list
        A list of tuples where the tuples are pairs of duplicates
    """

    duplicate_pairs = []  # define an empty list to be appended to
    for i, borehole_1 in enumerate(borefield):
        for j in range(i, len(borefield)):  # only loop unique interactions
            borehole_2 = borefield[j]
            if i == j:  # skip the borehole itself
                continue
            else:
                dist = sq_dist(borehole_1, borehole_2)
            if abs(dist) < (space * 10 ** -1):
                duplicate_pairs.append((i, j))
    if disp:
        # pad with '-' align in center
        output = f"{'*gt.boreholes.find_duplicates()*' :-^50}"
        # keep a space between the function name
        print(output.replace("*", " "))
        print(f"The duplicate pairs of boreholes found: {duplicate_pairs}")
    return duplicate_pairs


def sq_dist(p1, p2):
    """Returns the cartesian distance between two points"""
    return sqrt((p1[0] - p2[0]) * (p1[0] - p2[0]) + (p1[1] - p2[1]) * (p1[1] - p2[1]))


def remove_duplicates(borefield, space, disp=False):
    """
    Removes all the duplicates found from the duplicate pairs returned in
    :func:`check_duplicates`.
    For each pair of duplicates, the first borehole (with the lower index) is
    kept and the other (with the higher index) is removed.
    Parameters
    ----------
    borefield : list
        A list of :class:`Borehole` objects
    space:
    disp : bool, optional
        Set to true to print progression messages.
        Default is False.
    Returns
    -------
    new_borefield : list
        A boreField without duplicates
    """
    # get a list of tuple
    duplicate_pairs = find_duplicates(borefield, space, disp=disp)

    new_borefield = []

    # values not to be included
    duplicate_bores = []
    for i in range(len(duplicate_pairs)):
        duplicate_bores.append(duplicate_pairs[i][1])

    for i in range(len(borefield)):
        if i in duplicate_bores:
            continue
        else:
            new_borefield.append(borefield[i])
    if disp:
        # pad with '-' align in center
        print(
            f"{'*gt.boreholes.remove_duplicates()*' :-^50}".replace("*", " ")
        )  # keep a space between the function name
        n_duplicates = len(borefield) - len(new_borefield)
        print(f"The number of duplicates removed: {n_duplicates}")

    return new_borefield


def two_space_gen_bhc(
        field,
        y_space,
        x_space,
        no_go=None,
        rotate=0,
        p_space=None,
        i_space=None,
        intersection_tolerance=1e-5,
):
    """Generates a borefield that has perimeter spacing

    Parameters:
        field: The outer boundary of the property represented as an array of points
        y_space: Target Spacing in y-dir
        x_space: Target Spacing in x-dir
        no_go: a 3d array representing all the areas where boreholes cannot be placed
        rotate: the amount of rotation (rad)
        p_space: Perimeter spacing
        i_space: Spacing required between perimeter and the rest
        i_space: Min spacing required from all edges
        intersection_tolerance:

    """
    if p_space is None:
        p_space = 0.9 * x_space
    if i_space is None:
        i_space = x_space

    # calls the standard row-wise coord generator w/ the adjusted vertices
    holes = gen_borehole_config(
        field,
        y_space,
        x_space,
        no_go=no_go,
        rotate=rotate,
        intersection_tolerance=intersection_tolerance,
    )

    holes = holes.tolist()
    remove_points_too_close(field, holes, i_space, no_go_zones=no_go)

    # places the boreholes along the perimeter of the property boundary and no_go zone(s)
    perimeter_distribute(field, p_space, holes)
    if no_go is not None:
        for ng in no_go:
            perimeter_distribute(ng, p_space, holes)

    # returns the Holes as a numpy array for easier manipulation
    return_array = np.array(holes)
    return return_array


def remove_points_too_close(field, holes, i_space, no_go_zones=None):
    """
    Will remove all points too close to the field and no-go zones

    Parameters:
        field: The outer boundary of the property represented as an array of points
        no_go_zones: a 3d array representing all the areas where boreholes cannot be placed
        holes: 2d array containing all the current boreholes
        i_space: Min spacing required from all edges
    """
    field = field.c
    len_field = len(field)
    for i in range(len_field):
        p1 = field[i]
        if i == len_field - 1:
            p2 = field[0]
        else:
            p2 = field[i + 1]
        remove_points_close_too_line(p1, p2, holes, i_space)
    if no_go_zones is not None:
        len_no_go_zones = len(no_go_zones)
        for i in range(len_no_go_zones):
            ng = no_go_zones[i].c
            len_ng = len(ng)
            for j in range(len_ng):
                p1 = ng[j]
                if j == len_ng - 1:
                    p2 = ng[0]
                else:
                    p2 = ng[j + 1]
                remove_points_close_too_line(p1, p2, holes, i_space)


def remove_points_close_too_line(p1, p2, holes, i_space):
    """Removes points that are close to the given line

    Parameters:
        p1([float,float]): first point in line
        p2([float,float]): second point in line
        holes: 2d array containing a bunch of points
        i_space(float): distance cutoff for how close points can be

    """
    len_holes = len(holes)
    i = 0
    while i < len_holes:
        hole = holes[i]
        dp = dist_from_line(p1, p2, hole)
        if dp < i_space:
            del holes[i]
            len_holes -= 1
        else:
            i += 1


def dist_from_line(p1, p2, other_point):
    """Calculates the distance from a point to a line (closest distance):
    https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line

    Parameter:
        p1: first point on the line
        p2: second point on the line
        other_point: point which is being measured to

    """
    dxl = p2[0] - p1[0]
    dyl = p2[1] - p1[1]
    dx = p1[0] - other_point[0]
    dy = p1[1] - other_point[1]
    num = abs(dxl * dy - dx * dyl)
    den = sqrt(dxl * dxl + dyl * dyl)
    dp = num / den
    dist_l = sq_dist(p1, p2)
    d01 = sq_dist(p1, other_point)
    d02 = sq_dist(p2, other_point)
    if d01 * d01 - dp * dp < 0:
        return d01
    if sqrt(d01 * d01 - dp * dp) / dist_l > 1:
        return min(d01, d02)
    if min(p1[0], p2[0]) < other_point[0] < max(p1[0], p2[0]) or min(p1[1], p2[1]) < other_point[1] < max(p1[1], p2[1]):
        return min(d01, d02, dp)
    else:
        return min(d01, d02)


def perimeter_distribute(field, space, r):
    """
    Distributes boreholes along the perimeter of a given shape

    Parameters:
        field: array of points representing closed polygon
        space (float): spacing that the boreholes should have from one another
        r (dict{}): existing dictionary of boreholes which will be appended to
    """
    # print(r)
    for i in range(len(field.c)):
        if i == len(field.c) - 1:
            vert1 = field.c[i]
            vert2 = field.c[0]
        else:
            vert1 = field.c[i]
            vert2 = field.c[i + 1]
        dx = vert2[0] - vert1[0]
        dy = vert2[1] - vert1[1]

        # Checking how many boreholes can be distributed along the line
        dist = sq_dist(vert1, vert2)
        num_holes = int(dist // space)

        # Distributing the spacing to the x and y directions
        x_space = None
        y_space = None
        if num_holes > 0:
            x_space = dx / num_holes
            y_space = dy / num_holes

        current_p = [vert1[0], vert1[1]]

        # for loop is tuned to leave one spot empty for the next line
        for _ in range(num_holes):
            r.append([current_p[0], current_p[1]])
            current_p[0] += x_space
            current_p[1] += y_space


def gen_borehole_config(
        field,
        y_space,
        x_space,
        no_go=None,
        rotate=0,
        intersection_tolerance=1e-6,
):
    """
    Function generates a series of x,y points representing a field of boreholes
    in a trapezoidal shape. Returns empty if boreHole field does not meet given requirements

    Parameters
    -------------
        :param field:
        :param y_space: float
            the minimum spacing between points in the y-dir
        :param x_space: float
            the minimum spacing between points in the x-dir
        :param no_go:
        :param rotate:
        :param intersection_tolerance:

        :return: [[float]] -> 2 col + n rows
    """

    if no_go is None:
        no_go = []
    # Decides which vertex to start generating boreholes at by finding the "lowest" vertex relative to a rotated x-axis
    lowest_vert_val = float("inf")
    highest_vert_val = float("-inf")
    lowest_vert = None
    highest_vert = None
    for vert in field.c:
        if vert[0] != 0:
            phi = atan(vert[1] / vert[0])
        else:
            phi = PI_OVER_2
        dist_vert = sqrt(vert[1] ** 2 + vert[0] ** 2)
        ref_angle = phi
        if phi > PI_OVER_2:
            if phi > pi:
                if phi > 3 * PI_OVER_2:
                    ref_angle = 2 * rotate + 3 * PI_OVER_2 - phi
                else:
                    ref_angle = 2 * rotate + pi - phi
            else:
                ref_angle = pi - phi + 2 * rotate
        yp = dist_vert * sin(ref_angle - rotate)
        if yp < lowest_vert_val:
            lowest_vert_val = yp
            lowest_vert = vert
        if yp > highest_vert_val:
            highest_vert_val = yp
            highest_vert = vert

    # Determines the number of rows as well as the distance between the rows
    num_rows = int((highest_vert_val - lowest_vert_val) // y_space)
    d = highest_vert_val - lowest_vert_val
    s = d / num_rows
    row_space = [-1 * s * cos(PI_OVER_2 - rotate), s * sin(PI_OVER_2 - rotate)]

    # Establishes the dictionary where the boreholes will be added two as well as establishing a point on the first row
    boreholes = {}
    row_point = [lowest_vert[0], lowest_vert[1]]

    # This is just a value that is combined with the slope of the row's to establish two points defining a row (could
    # be any value)
    point_shift = 1000.0

    for _ in range(num_rows + 1):

        # Row Defined by two points
        if row_space[1] == 0:
            row = [
                row_point[0],
                row_point[1],
                row_point[0],
                row_point[1] + point_shift,
            ]
        else:
            row = [
                row_point[0],
                row_point[1],
                row_point[0] + point_shift,
                row_point[1] + (-row_space[0] / row_space[1]) * point_shift,
            ]

        # Gets Intersection between current row and property boundary
        f_inters = field.line_intersect(row, rotate, intersection_tolerance)

        # Stores the number of intersections with the row
        len_f_inters = len(f_inters)

        # Checks for edge case where a single intersection is reported as two and treats it as one
        if (
                len_f_inters > 1
                and abs(f_inters[0][0] - f_inters[1][0]) <= intersection_tolerance
                and abs(f_inters[0][1] - f_inters[1][1]) <= intersection_tolerance
        ):
            fi = 0
            fij = 0
            while fi < len_f_inters:
                while fij < len_f_inters:
                    if fij == fi:
                        fij += 1
                        continue
                    if (
                            abs(f_inters[fi][0] - f_inters[fij][0]) <= intersection_tolerance
                            and abs(f_inters[fi][1] - f_inters[fij][1])
                            <= intersection_tolerance
                    ):
                        f_inters.pop(fij)
                        if fi >= fij:
                            fi -= 1
                        fij -= 1
                        len_f_inters -= 1
                    fij += 1
                fi += 1

        # Checks for edge case where there are no intersections detected due to a rounding error (can sometimes
        # happen with the last row)
        """

        if len_f_inters == 0 and ri == num_rows:
            ins = False

            # Checks if the predicted point (ghost point that was expected but not found) is inside one of the no_go
            zones for shape in no_go: if shape.point_intersect(highest_vert): ins = True if not ins: #Double checks
            that this borehole has not already been included if len(boreholes)==0 or not (boreholes[len(boreholes) -
            1][0] == highest_vert[0] and boreholes[len(boreholes) - 1][1] ==highest_vert[1]): boreholes[len(
            boreholes)] = highest_vert """
        # Handles cases with odd number of intersections
        if len_f_inters % 2 == 0:

            # Specific case with two intersections
            if len_f_inters == 2:

                # Checks for the edge case where two intersections are very close together and replaces them with one
                # point
                if (
                        sqrt(
                            (f_inters[0][0] - f_inters[1][0])
                            * (f_inters[0][0] - f_inters[1][0])
                            + (f_inters[0][1] - f_inters[1][1])
                            * (f_inters[0][1] - f_inters[1][1])
                        )
                        < x_space
                ):
                    ins = False
                    for ng_shape in no_go:
                        if ng_shape.point_intersect(highest_vert):
                            ins = True
                    if not ins:
                        boreholes[len(boreholes)] = f_inters[0]
                        len_f_inters = 0  # skips the while loop

            i = 0
            while i < len_f_inters - 1:

                left_offset = [0, 0]
                right_offset = [0, 0]

                # Checks if there is enough distance between this point and another and then will offset the point if
                # there is not enough room
                dls_check = sqrt(
                    (f_inters[i][0] - f_inters[i - 1][0])
                    * (f_inters[i][0] - f_inters[i - 1][0])
                    + (f_inters[i][1] - f_inters[i - 1][1])
                    * (f_inters[i][1] - f_inters[i - 1][1])
                )

                drs_check = sqrt(
                    (f_inters[i][0] - f_inters[i + 1][0])
                    * (f_inters[i][0] - f_inters[i + 1][0])
                    + (f_inters[i][1] - f_inters[i + 1][1])
                    * (f_inters[i][1] - f_inters[i + 1][1])
                )

                if i > 0 and (dls := dls_check) < x_space:
                    left_offset = [dls * cos(rotate), dls * sin(rotate)]
                elif i < len_f_inters - 1 and (drs := drs_check) < x_space:
                    right_offset = [-drs * cos(rotate), -drs * sin(rotate)]

                process_rows(
                    row,
                    [f_inters[i][0] + left_offset[0], f_inters[i][1] + left_offset[1]],
                    [
                        f_inters[i + 1][0] + right_offset[0],
                        f_inters[i + 1][1] + right_offset[1],
                    ],
                    no_go,
                    x_space,
                    boreholes,
                    rotate=rotate,
                )

                i += 2
        elif len_f_inters == 1:
            ins = False
            for ng_shape in no_go:
                if ng_shape.point_intersect(highest_vert):
                    ins = True
            if not ins:
                if len(boreholes) == 0:
                    boreholes[len(boreholes)] = f_inters[0]
                if not (
                        boreholes[len(boreholes) - 1][0] == f_inters[0][0]
                        and boreholes[len(boreholes) - 1][1] == f_inters[0][1]
                ):
                    boreholes[len(boreholes)] = f_inters[0]
        else:
            i = 0
            while i < len_f_inters - 1:
                if field.point_intersect(
                        [
                            (f_inters[i][0] + f_inters[i + 1][0]) / 2,
                            (f_inters[i][1] + f_inters[i + 1][1]) / 2,
                        ]
                ):
                    process_rows(
                        row,
                        f_inters[i],
                        f_inters[i + 1],
                        no_go,
                        x_space,
                        boreholes,
                        rotate=rotate,
                    )
                i += 1
        row_point[0] += row_space[0]
        row_point[1] += row_space[1]
    r_a = [boreholes[element] for element in boreholes]
    r_a = np.array(remove_duplicates(r_a, x_space))
    return r_a


def process_rows(row, row_sx, row_ex, no_go, row_space, r_a, rotate, intersection_tolerance=1e-5):
    """
    Function generates a row of the borefield
    *Note: the formatting from the rows can be a little unexpected. Some adjustment
    may be required to correct the formatting. The genBoreHoleConfig function already accounts for this.
    Parameters
    -------------
    :param row:
    :param row_sx:
    :param row_ex:
    :param no_go:
    :param row_space:
    :param r_a:
    :param rotate:
    :param intersection_tolerance:
    """

    if no_go is None:
        distribute(row_sx, row_ex, row_space, r_a, rotate)
        return r_a
    num_col = int(
        sqrt(
            (row_sx[0] - row_ex[0]) * (row_sx[0] - row_ex[0])
            + (row_sx[1] - row_ex[1]) * (row_sx[1] - row_ex[1])
        )
        // row_space
    )

    inters = [
        point
        for shape in no_go
        for point in shape.line_intersect(
            row, rotate=rotate, intersection_tolerance=intersection_tolerance
        )
    ]
    inters = sort_intersections(inters, rotate)
    num_inters = len(inters)

    if num_inters > 1:
        if less_than(
                inters[0],
                row_sx,
                rotate=rotate,
                intersection_tolerance=intersection_tolerance,
        ) and less_than(
            row_ex,
            inters[len(inters) - 1],
            rotate=rotate,
            intersection_tolerance=intersection_tolerance,
        ):
            inside = False
            for _ in inters:
                if less_than(
                        row_sx,
                        inters[0],
                        rotate=rotate,
                        intersection_tolerance=intersection_tolerance,
                ) and less_than(
                    inters[0],
                    row_ex,
                    rotate=rotate,
                    intersection_tolerance=intersection_tolerance,
                ):
                    inside = True
            if not inside:
                point_in = False
                for ngShape in no_go:
                    if ngShape.point_intersect(
                            [(row_ex[0] + row_sx[0]) / 2, (row_ex[1] + row_sx[1]) / 2]
                    ):
                        point_in = True
                if point_in:
                    return []
    inters = np.array(inters)
    indices = []
    for j in range(num_inters):
        less_than_1 = less_than(row_ex, inters[j], rotate=rotate, intersection_tolerance=intersection_tolerance)
        less_than_2 = less_than(inters[j], row_sx, rotate=rotate, intersection_tolerance=intersection_tolerance)
        if not (less_than_1 or less_than_2):
            indices.append(j)
    inters = inters[indices]
    num_inters = len(inters)
    for i in range(num_inters - 1):
        space = sqrt((inters[i + 1][0] - inters[i][0]) * (inters[i + 1][0] - inters[i][0])
                     + (inters[i + 1][1] - inters[i][1]) * (inters[i + 1][1] - inters[i][1]))
        if space < row_space:
            i_none = False
            for shape in no_go:
                if shape.point_intersect(
                        [
                            (inters[i + 1][0] + inters[i][0]) / 2,
                            (inters[i + 1][1] + inters[i][1]) / 2,
                        ]
                ):
                    i_none = True
            if i_none:
                d = (row_space - space) / 2
                inters[i + 1][0] += d * cos(rotate)
                inters[i + 1][1] += d * sin(rotate)
                inters[i][0] -= d * cos(rotate)
                inters[i][1] -= d * sin(rotate)
    if num_col < 1:
        ins = False
        for shape in no_go:
            if shape.point_intersect(
                    [(row_ex[0] + row_sx[0]) / 2, (row_ex[1] + row_sx[1]) / 2]
            ):
                ins = True
        if not ins:
            if len(r_a) == 0 or not (
                    r_a[len(r_a) - 1][0] == (row_ex[0] + row_sx[0]) / 2
                    and r_a[len(r_a) - 1][1] == (row_ex[1] + row_sx[1]) / 2
            ):
                r_a[len(r_a)] = [(row_ex[0] + row_sx[0]) / 2, (row_ex[1] + row_sx[1]) / 2]
            return r_a
    else:
        if num_inters == 0:
            if not_inside(row_sx, no_go) and not_inside(row_ex, no_go):
                distribute(row_sx, row_ex, row_space, r_a, rotate)
        elif num_inters == 2:
            distribute(row_sx, inters[0], row_space, r_a, rotate)
            distribute(inters[1], row_ex, row_space, r_a, rotate)
        elif num_inters == 1:
            ins = False
            for shape in no_go:
                if shape.point_intersect(
                        [(inters[0][0] + row_sx[0]) / 2, (inters[0][1] + row_sx[1]) / 2]
                ):
                    ins = True
            if not ins:
                distribute(row_sx, inters[0], row_space, r_a, rotate)
            else:
                distribute(inters[0], row_ex, row_space, r_a, rotate)
        elif num_inters % 2 == 0:
            i = 0
            while i < num_inters:
                if i == 0:
                    distribute(row_sx, inters[0], row_space, r_a, rotate)
                    i = 1
                    continue
                elif i == num_inters - 1:
                    distribute(inters[num_inters - 1], row_ex, row_space, r_a, rotate)
                else:
                    distribute(inters[i], inters[i + 1], row_space, r_a, rotate)
                i += 2
        else:
            ins = False
            for shape in no_go:
                if shape.point_intersect(
                        [(inters[0][0] + row_sx[0]) / 2, (inters[0][1] + row_sx[1]) / 2]
                ):
                    ins = True
            if not ins:
                i = 0
                while i < num_inters:
                    if i == 0:
                        distribute(row_sx, inters[0], row_space, r_a, rotate)
                        i = 1
                        continue
                    elif i == num_inters - 1:
                        i += 2
                        continue
                    else:
                        distribute(inters[i], inters[i + 1], row_space, r_a, rotate)
                    i += 2
            else:
                i = 0
                while i < num_inters:
                    if i == 0:
                        distribute(inters[0], inters[1], row_space, r_a, rotate)
                        i = 2
                        continue
                    elif i == num_inters - 1:
                        distribute(inters[i], row_ex, row_space, r_a, rotate)
                        i += 2
                        continue
                    else:
                        distribute(inters[i], inters[i + 1], row_space, r_a, rotate)
                    i += 2

    return r_a


def not_inside(p, ngs):
    inside = False
    for ng in ngs:
        if ng.point_intersect(p):
            inside = True
    return not inside


def less_than(p1, p2, rotate=0, intersection_tolerance=1e-5):
    x1, y1 = p1
    x2, y2 = p2
    dx = x2 - x1
    dy = y2 - y1

    if abs(dx) < intersection_tolerance:
        dx_sign = 0
    elif dx > 0:
        dx_sign = 1
    else:
        dx_sign = -1

    if abs(dy) < intersection_tolerance:
        dy_sign = 0
    elif dy > 0:
        dy_sign = 1
    else:
        dy_sign = -1

    if rotate >= 0:
        if dx_sign == 0:
            if dy_sign == 1:
                return True
            else:
                return False
        elif dy_sign == 0:
            if dx_sign == 1:
                return True
            else:
                return False
        elif dx_sign == dy_sign:
            if dx_sign == 1:
                return True
            else:
                return False
        else:
            raise ValueError("Slope between points does not match field orientation.")
    else:
        if dx_sign == 0:
            if dy_sign == 1:
                return False
            else:
                return False
        elif dy_sign == 0:
            if dx_sign == 1:
                return True
            else:
                return False
        elif dx_sign != dy_sign:
            if dx_sign == 1:
                return True
            else:
                return False
        else:
            raise ValueError("Slope between points does not match field orientation.")


def distribute(x1, x2, spacing, r, rotate):
    """
      Function generates a series of boreholes between x1 and x2
    Parameters
    -------------
    :param x1: float
        left x value
    :param x2: float
        right x value
    :param spacing: float
        spacing between columns
    :param r: [[float]]
        existing array of points
    :param rotate:
    :return:
    """
    dx = sqrt((x1[0] - x2[0]) * (x1[0] - x2[0]) + (x1[1] - x2[1]) * (x1[1] - x2[1]))
    if dx < spacing:
        if len(r) == 0 or not (
                r[len(r) - 1][0] == (x1[0] + x2[0]) / 2
                and r[len(r) - 1][1] == (x1[1] + x2[1]) / 2
        ):
            r[len(r)] = [(x1[0] + x2[0]) / 2, (x1[1] + x2[1]) / 2]
        return
    current_x = x1
    act_num_col = int(dx // spacing)
    act_space = dx / act_num_col
    while (
            sqrt(
                (current_x[0] - x2[0]) * (current_x[0] - x2[0])
                + (current_x[1] - x2[1]) * (current_x[1] - x2[1])
            )
    ) >= 1e-8:
        if len(r) == 0 or not (
                r[len(r) - 1][0] == current_x[0] and r[len(r) - 1][1] == current_x[1]
        ):
            r[len(r)] = [current_x[0], current_x[1]]
        current_x[0] += act_space * cos(rotate)
        current_x[1] += act_space * sin(rotate)
    if not (r[len(r) - 1][0] == x2[0] and r[len(r) - 1][1] == x2[1]):
        r[len(r)] = [x2[0], x2[1]]
    return



from ghedesigner.shape import point_polygon_check


def remove_cutout(coordinates, boundary, remove_inside=True, keep_contour=True):
    inside_points_idx = []
    outside_points_idx = []
    boundary_points_idx = []

    inside = 1
    outside = -1
    on_edge = 0

    for idx, coordinate in enumerate(coordinates):
        coordinate = coordinates[idx]
        ret = point_polygon_check(boundary, coordinate)
        if ret == inside:
            inside_points_idx.append(idx)
        elif ret == outside:
            outside_points_idx.append(idx)
        elif ret == on_edge:
            boundary_points_idx.append(idx)
        else:
            raise ValueError("Something bad happened")

    new_coordinates = []
    for idx, _ in enumerate(coordinates):
        # if we want to remove inside points and keep contour points
        if remove_inside and keep_contour:
            if idx in inside_points_idx:
                continue
            else:
                new_coordinates.append(coordinates[idx])
        # if we want to remove inside points and remove contour points
        elif remove_inside and not keep_contour:
            if idx in inside_points_idx or idx in boundary_points_idx:
                continue
            else:
                new_coordinates.append(coordinates[idx])
        # if we want to keep outside points and remove contour points
        elif not remove_inside and not keep_contour:
            if idx in outside_points_idx or idx in boundary_points_idx:
                continue
            else:
                new_coordinates.append(coordinates[idx])
        # if we want to keep outside points and keep contour points
        else:
            if idx in outside_points_idx:
                continue
            else:
                new_coordinates.append(coordinates[idx])

    return new_coordinates


def determine_largest_rectangle(property_boundary):
    x_max = 0
    y_max = 0
    for x, y in property_boundary:
        if x > x_max:
            x_max = x
        if y > y_max:
            y_max = y

    rectangle = [[0, 0], [x_max, 0], [x_max, y_max], [0, y_max], [0, 0]]

    return rectangle



import warnings
from math import log

import numpy as np
import pygfunction as gt
from scipy.interpolate import interp1d, lagrange

from ghedesigner.borehole import GHEBorehole
from ghedesigner.borehole_heat_exchangers import get_bhe_object
from ghedesigner.enums import BHPipeType


def calculate_g_function(
        m_flow_borehole,
        bhe_type: BHPipeType,
        time_values,
        coordinates,
        borehole,
        fluid,
        pipe,
        grout,
        soil,
        n_segments=8,
        end_length_ratio=0.02,
        segments="unequal",
        solver="equivalent",
        boundary="MIFT",
        segment_ratios=None,
        disp=False,
):
    bore_field = []
    bhe_objects = []

    h = borehole.H
    r_b = borehole.r_b
    d = borehole.D
    tilt = borehole.tilt
    orientation = borehole.orientation

    for x, y in coordinates:
        _borehole = GHEBorehole(h, d, r_b, x, y, tilt, orientation)
        bore_field.append(_borehole)
        # Initialize pipe model
        if boundary == "MIFT":
            bhe = get_bhe_object(bhe_type, m_flow_borehole, fluid, _borehole, pipe, grout, soil)
            bhe_objects.append(bhe)

    alpha = soil.k / soil.rhoCp

    # setup options
    segments = segments.lower()
    if segments == "equal":
        options = {"nSegments": n_segments, "disp": disp}
    elif segments == "unequal":
        if segment_ratios is None:
            segment_ratios = gt.utilities.segment_ratios(n_segments, end_length_ratio=end_length_ratio)
        else:
            segment_ratios = segment_ratios
        options = {
            "nSegments": n_segments,
            "segment_ratios": segment_ratios,
            "disp": disp,
        }
    else:
        raise ValueError("Equal or Unequal are acceptable options " "for segments.")

    if boundary == "UHTR" or boundary == "UBWT":
        gfunc = gt.gfunction.gFunction(
            bore_field,
            alpha,
            time=time_values,
            boundary_condition=boundary,
            options=options,
            method=solver,
        )
    elif boundary == "MIFT":
        m_flow_network = len(bore_field) * m_flow_borehole
        network = gt.networks.Network(bore_field, bhe_objects, m_flow_network=m_flow_network, cp_f=fluid.cp)
        gfunc = gt.gfunction.gFunction(
            network,
            alpha,
            time=time_values,
            boundary_condition=boundary,
            options=options,
            method=solver,
        )
    else:
        raise ValueError("UHTR, UBWT or MIFT are accepted boundary conditions.")

    return gfunc


def calc_g_func_for_multiple_lengths(
        b: float,
        h_values: list,
        r_b,
        depth,
        m_flow_borehole,
        bhe_type: BHPipeType,
        log_time,
        coordinates,
        fluid,
        pipe,
        grout,
        soil,
        n_segments=8,
        segments="unequal",
        solver="equivalent",
        boundary="MIFT",
        segment_ratios=None,
):
    d = {"g": {}, "bore_locations": coordinates, "logtime": log_time}

    for h in h_values:
        _borehole = GHEBorehole(h, depth, r_b, 0.0, 0.0)

        alpha = soil.k / soil.rhoCp

        ts = h ** 2 / (9.0 * alpha)  # Bore field characteristic time
        time_values = np.exp(log_time) * ts

        gfunc = calculate_g_function(
            m_flow_borehole,
            bhe_type,
            time_values,
            coordinates,
            _borehole,
            fluid,
            pipe,
            grout,
            soil,
            n_segments=n_segments,
            segments=segments,
            solver=solver,
            boundary=boundary,
            segment_ratios=segment_ratios,
        )

        key = f"{b}_{h}_{r_b}_{d}"

        d["g"][key] = gfunc.gFunc.tolist()

    geothermal_g_input = GFunction.configure_database_file_for_usage(d)
    # Initialize the gFunction object
    g_function = GFunction(**geothermal_g_input)

    return g_function


class GFunction:
    def __init__(
            self,
            b: float,
            r_b_values: dict,
            d_values: dict,
            g_lts: dict,
            log_time: list,
            bore_locations: list,
    ):
        self.B: float = b  # a B spacing in the borefield
        # r_b (borehole radius) value keyed by height
        self.r_b_values: dict = r_b_values
        self.D_values: dict = d_values  # D (burial depth) value keyed by height
        self.g_lts: dict = g_lts  # g-functions (LTS) keyed by height
        # ln(t/ts) values that apply to all the heights
        self.log_time: list = log_time
        # (x, y) coordinates of boreholes
        self.bore_locations: list = bore_locations
        # self.time: dict = {}  # the time values in years

        # an interpolation table for B/H ratios, D, r_b (used in the method
        # g_function_interpolation)
        self.interpolation_table: dict = {}

    def g_function_interpolation(self, b_over_h: float, kind="default"):

        # the g-functions are stored in a dictionary based on heights, so an
        # equivalent height can be found
        h_eq = 1 / b_over_h * self.B

        # Determine if we are out of range and need to extrapolate
        height_values = list(self.g_lts.keys())
        # If we are close to the outer bounds, then set H_eq as outer bounds
        if abs(h_eq - max(height_values)) < 1.0e-6:
            h_eq = max(height_values)
        if abs(h_eq - min(height_values)) < 1.0e-6:
            h_eq = min(height_values)

        if min(height_values) <= h_eq <= max(height_values):
            fill_value = ""
        elif abs(min(height_values) - h_eq) < 0.001:
            fill_value = ""
        else:
            fill_value = "extrapolate"
            warnings.warn("Extrapolation is being used.")

        # if the interpolation kind is default, use what we know about the
        # accuracy of interpolation to choose a technique
        if kind == "default":
            num_curves = len(height_values)
            if num_curves >= 5:
                kind = "cubic"
            elif num_curves >= 3:
                kind = "quadratic"
            elif num_curves == 2:
                kind = "linear"
            else:
                if (h_eq - height_values[0]) / height_values[0] < 0.001:
                    g_function = self.g_lts[height_values[0]]
                    rb = self.r_b_values[height_values[0]]
                    d = self.D_values[height_values[0]]
                    return g_function, rb, d, h_eq
                elif min(height_values) - h_eq < 0.001:
                    g_function = self.g_lts[height_values[0]]
                    rb = self.r_b_values[height_values[0]]
                    d = self.D_values[height_values[0]]
                    return g_function, rb, d, h_eq
                else:
                    raise ValueError(
                        "The interpolation requires two g-function curves "
                        "if the requested B/H is not already computed.")

        # Automatically adjust interpolation if necessary
        # Lagrange also needs 2
        interpolation_kinds = {"linear": 2, "quadratic": 3, "cubic": 4, "lagrange": 2}
        curves_by_kind = {2: "linear", 3: "quadratic", 4: "cubic"}
        if len(height_values) < 2:
            raise ValueError("Interpolation requires two g-function curves.")
        else:
            # If the number of required curves for the interpolation type is
            # greater than what is available, reduce the interpolation type
            required_curves = interpolation_kinds[kind]
            if required_curves > len(height_values):
                kind = curves_by_kind[len(height_values)]

        # if the interpolation table is not yet know, build it
        if len(self.interpolation_table) == 0:
            # create an interpolation for the g-function which takes the height
            # (or equivalent height) as an input the g-function needs to be
            # interpolated at each point in dimensionless time
            self.interpolation_table["g"] = []
            for i, _ in enumerate(self.log_time):
                x = []
                y = []
                for key in self.g_lts:
                    height_value = float(key)
                    g_value = self.g_lts[key][i]
                    x.append(height_value)
                    y.append(g_value)
                if kind == "lagrange":
                    f = lagrange(x, y)
                else:
                    f = interp1d(x, y, kind=kind, fill_value=fill_value)
                self.interpolation_table["g"].append(f)
            # create interpolation tables for 'D' and 'r_b' by height
            keys = list(self.r_b_values.keys())
            height_values: list = []
            rb_values: list = []
            d_values: list = []
            for h in keys:
                height_values.append(float(h))
                rb_values.append(self.r_b_values[h])
                try:
                    d_values.append(self.D_values[h])
                except:
                    pass
            if kind == "lagrange":
                rb_f = lagrange(height_values, rb_values)
            else:
                # interpolation function for rb values by H equivalent
                rb_f = interp1d(height_values, rb_values, kind=kind, fill_value=fill_value)
            self.interpolation_table["rb"] = rb_f
            try:
                if kind == "lagrange":
                    d_f = lagrange(height_values, d_values)
                else:
                    d_f = interp1d(height_values, d_values, kind=kind, fill_value=fill_value)
                self.interpolation_table["D"] = d_f
            except:
                pass

        # create the g-function by interpolating at each ln(t/ts) value
        rb_value = self.interpolation_table["rb"](h_eq)
        try:
            d_value = self.interpolation_table["D"](h_eq)
        except:
            d_value = None
        g_function: list = []
        for i in range(len(self.log_time)):
            f = self.interpolation_table["g"][i]
            g = f(h_eq).tolist()
            g_function.append(g)
        return g_function, rb_value, d_value, h_eq

    @staticmethod
    def borehole_radius_correction(g_function: list, rb: float, rb_star: float):
        r"""
        Correct the borehole radius. From paper 3 of Eskilson 1987.
        .. math::
            g(\dfrac{t}{t_s}, \dfrac{r_b^*}{H}) =
            g(\dfrac{t}{t_s}, \dfrac{r_b}{H}) - ln(\dfrac{r_b^*}{r_b})
        Parameters
        ----------
        g_function: list
            A g-function
        rb: float
            The current borehole radius
        rb_star: float
            The borehole radius that is being corrected to
        Returns
        -------
        g_function_corrected: list
            A corrected g_function
        """
        g_function_corrected = []
        for g in g_function:
            g_function_corrected.append(g - log(rb_star / rb))
        return g_function_corrected

    @staticmethod
    def configure_database_file_for_usage(data) -> dict:
        """
        This method is called upon initialization of the object.
        Read the cpgfunction output dictionary into the borefield class for easy
        access of information
        Parameters
        ----------
        data: dict
            A dictionary which is in the output format of cpgfunction.
        Returns
        -------
            a dictionary for input to this object
        """
        log_time = data["logtime"]

        bore_locations = data[
            "bore_locations"
        ]  # store the bore locations in the object
        g_values: dict = data["g"]  # pull the g-functions into the g_values
        # a temporary g-function dictionary that might be out of order
        g_tmp: dict = {}
        # a temporary burial depth dictionary that may be out of order
        ds_tmp: dict = {}
        # the borehole radius dictionary that may be out of order
        r_bs_tmp: dict = {}

        for key in g_values:
            # do the g-function dictionary
            key_split = key.split("_")
            # get the current height
            height = float((key_split[1]))
            # create a g-function list associated with this height key
            g_tmp[height] = g_values[key]
            # create an r_b value associated with this height key
            r_b = float(key_split[2])
            r_bs_tmp[height] = r_b
            # the D value is recently added to the key value for the saved
            # g-functions computed
            try:
                d = float(key_split[3])
                ds_tmp[height] = d
            except:
                pass

        # every B-spacing should be the same for each file
        b = float(list(g_values.keys())[0].split("_")[0])

        keys = sorted(list(g_tmp.keys()), key=int)  # sort the heights in order
        # fill the g-function dictionary with sorted heights
        g = {key: g_tmp[key] for key in keys}
        # fill the burial depth dictionary with sorted heights
        try:
            ds = {key: ds_tmp[key] for key in keys}
        except:
            # if there's no D provided, make it 2
            ds = {key: 2.0 for key in keys}
        r_bs = {key: r_bs_tmp[key] for key in keys}

        geothermal_g_input = {
            "b": b,
            "r_b_values": r_bs,
            "d_values": ds,
            "g_lts": g,
            "log_time": log_time,
            "bore_locations": bore_locations
        }

        return geothermal_g_input



from math import atan, pi, sin, sqrt

import numpy as np

from ghedesigner.constants import PI_OVER_2, TWO_PI


class Shapes:

    def __init__(self, c):
        """
         constructs a shape object
        """
        self.c = np.array(c)
        # print(c)
        xs = [0] * len(self.c)
        ys = [0] * len(self.c)
        for i in range(len(self.c)):
            xs[i] = self.c[i][0]
            ys[i] = self.c[i][1]
        self.max_x = max(xs)
        self.min_x = min(xs)
        self.max_y = max(ys)
        self.min_y = min(ys)

    def line_intersect(self, xy, rotate=0, intersection_tolerance=1e-6):
        """
        returns the intersections between a line segment and the shape

        Parameters
        -----------
        :param xy: [float,float,float,float]
            the x,y values of both endpoints of the line segment
        :param rotate:
        :param intersection_tolerance:

        :return: [[float]]
            the x,y values of the intersections
        """
        x1, y1, x2, y2 = xy
        r_a = []
        for i in range(len(self.c)):
            if i == len(self.c) - 1:
                c1 = self.c[len(self.c) - 1]
                c2 = self.c[0]
                r = vector_intersect(
                    [c1[0], c1[1], c2[0], c2[1]],
                    [x1, y1, x2, y2],
                    intersection_tolerance,
                )
                # print(r)
                if len(r) == 1:
                    r = r[0]
                    if (
                            (r[0] - max(c2[0], c1[0])) > intersection_tolerance
                            or (r[0] - min(c2[0], c1[0])) < -1 * intersection_tolerance
                            or (r[1] - max(c2[1], c1[1])) > intersection_tolerance
                            or (r[1] - min(c2[1], c1[1])) < -1 * intersection_tolerance
                    ):
                        continue
                    r_a.append(r)
            else:
                c1 = self.c[i]
                c2 = self.c[i + 1]
                r = vector_intersect(
                    [c1[0], c1[1], c2[0], c2[1]],
                    [x1, y1, x2, y2],
                    intersection_tolerance,
                )
                if len(r) == 1:
                    r = r[0]
                    if (
                            (r[0] - max(c2[0], c1[0])) > intersection_tolerance
                            or (r[0] - min(c2[0], c1[0])) < -1 * intersection_tolerance
                            or (r[1] - max(c2[1], c1[1])) > intersection_tolerance
                            or (r[1] - min(c2[1], c1[1])) < -1 * intersection_tolerance
                    ):
                        continue
                    r_a.append(r)
        # print("x value: %f, r values:"%x1)
        # print(r_a)

        r_a = sort_intersections(r_a, rotate)
        # print(r_a)
        return r_a

    def point_intersect(self, xy):
        """
        returns whether the given point is inside the rectangle

        Parameters
        -----------
        :param xy: [float,float]
            x,y value of point

        :return: boolean
            true if inside, false if not
        """
        x, y = xy
        if (x > self.max_x or x < self.min_x) or (y > self.max_y or y < self.min_y):
            # print("Returning False b/c outside of box")
            return False
        far_x = self.min_x - 10
        inters = self.line_intersect([far_x, y, far_x + 1, y])
        # print(inters)
        inters = [inter for inter in inters if inter[0] <= x]
        # print("x: %f"%x,inters)
        if len(inters) == 1:
            # print("Returning True")
            return True
        i = 0
        while i < len(inters):
            for vert in self.c:
                if inters[i][0] == vert[0] and inters[i][1] == vert[1]:
                    inters.pop(i)
                    i -= 1
                    break
            i += 1
        if len(inters) % 2 == 0:
            # print(len(inters))
            return False
        else:
            # print("returning True")
            return True

    def get_area(self):
        """
        returns area of shape

        :return: float
            area of shape
        """
        area_sum = 0
        for i in range(len(self.c)):
            if i == len(self.c) - 1:
                area_sum += self.c[len(self.c) - 1][0] * self.c[0][1] - (self.c[len(self.c) - 1][1] * self.c[0][0])
                continue
            area_sum += self.c[i][0] * self.c[i + 1][1] - (self.c[i][1] * self.c[i + 1][0])
        return 0.5 * area_sum


def sort_intersections(r_a, rotate):
    if len(r_a) == 0:
        return r_a
    vals = [0] * len(r_a)
    i = 0
    for inter in r_a:
        if inter[0] == 0:
            phi = PI_OVER_2
        else:
            phi = atan(inter[1] / inter[0])
        dist_inter = sqrt(inter[1] ** 2 + inter[0] ** 2)
        ref_ang = PI_OVER_2 - phi
        # sign = 1
        if phi > PI_OVER_2:
            if phi > pi:
                if phi > 3 * PI_OVER_2:
                    ref_ang = TWO_PI - phi
                else:
                    ref_ang = 3.0 * PI_OVER_2 - phi
            else:
                ref_ang = pi - phi
        # if phi > pi/2 + rotate and phi < 3*pi/2 + rotate:
        # sign = -1
        vals[i] = dist_inter * sin(ref_ang + rotate)
        i += 1
    zipped = sorted(zip(vals, r_a))
    r_a = [row for _, row in zipped]
    return r_a


def vector_intersect(l1, l2, intersection_tolerance):
    """
     gives the intersection between two line segments

    Parameters
    -----------
    :param l1: [[float]]
        endpoints of first line segment
    :param l2: [[float]]
        endpoints of the second line segment
    :param intersection_tolerance:

    :return: [float,float]
        x,y values of intersection (returns None if there is none)
    """
    x11, y11, x12, y12 = l1
    x21, y21, x22, y22 = l2
    if x12 - x11 == 0:
        a1 = float("inf")
    else:
        a1 = (y12 - y11) / (x12 - x11)
        c1 = y11 - x11 * a1
    if x22 - x21 == 0:
        a2 = float("inf")
    else:
        a2 = (y22 - y21) / (x22 - x21)
        c2 = y21 - x21 * a2
    if a1 == float("inf") or a2 == float("inf"):
        if a1 == float("inf") and a2 == float("inf"):
            if abs(x11 - x21) < intersection_tolerance:
                return [[x11, y11], [x12, y12]]
            else:
                return []
        elif a1 == float("inf"):
            return [[x11, a2 * x11 + c2]]
        else:
            return [[x21, a1 * x21 + c1]]
    if abs(a1 - a2) <= intersection_tolerance:
        if abs(y22 - (a1 * x22 + c1)) <= intersection_tolerance:
            # return [[x11,y11],[x12,y12]]
            return []
        else:
            return []
    rx = (c2 - c1) / (a1 - a2)
    ry = a1 * (c2 - c1) / (a1 - a2) + c1
    return [[rx, ry]]


def point_polygon_check(contour, point):
    """
    Mimics pointPolygonTest from OpenCV-Python

    Adapted from the methods outlined in the links below.
    https://stackoverflow.com/a/63436180/5965685
    https://stackoverflow.com/a/17693146/5965685

    :param contour: list of tuples containing (x, y) contour boundary points
    :param point: tuple containing the (x, y) point to test

    :returns: -1 if outside, 0 if on edge, 1 if inside
    :rtype: int
    """

    # check if on edge
    # use Pythagoras to check whether the distance between the test point
    # and the line vertices add to the distance between the line vertices
    # if they are within tolerance, the point is co-linear
    # if not, it is off the line

    def distance(pt_1, pt_2) -> float:
        return sqrt((pt_1[0] - pt_2[0]) ** 2 + (pt_1[1] - pt_2[1]) ** 2)

    on_edge_tolerance = 0.001

    for idx, vertex in enumerate(contour):
        v1 = contour[idx - 1]
        v2 = vertex
        test_dist = distance(v1, point) + distance(v2, point)
        v12_dist = distance(v1, v2)

        if abs(test_dist - v12_dist) < on_edge_tolerance:
            return 0

    # if made it to here, not on edge and check if inside/outside
    def between(p, a, b) -> bool:
        return ((p >= a) and (p <= b)) or ((p <= a) and (p >= b))

    inside = True
    px = point[0]
    py = point[1]

    for idx, vertex in enumerate(contour):
        v1 = contour[idx - 1]
        v2 = vertex
        v1x = v1[0]
        v1y = v1[1]
        v2x = v2[0]
        v2y = v2[1]

        if between(py, v1y, v2y):  # points inside vertical range
            if ((py == v1y) and (v2y >= v1y)) or ((py == v2y) and (v1y >= v2y)):
                continue

            # calc cross product `PA X PB`, P lays on left side of AB if c > 0
            c = (v1x - px) * (v2y - py) - (v2x - px) * (v1y - py)

            if c == 0:
                return 0

            if (v1y < v2y) == (c > 0):
                inside = not inside

    return -1 if inside else 1



import sys
from json import loads
from pathlib import Path

from jsonschema import validate, ValidationError

from ghedesigner.enums import BHPipeType, DesignGeomType


# Note: JSON schema does not currently have a good way to handle case-insensitive enums.
#       Some fields are upper-cased manually here for validation purposes.
#       More details here: https://github.com/json-schema-org/community/discussions/148


def validate_schema_instance(schema_file_name: str, instance: dict, error_msg: str) -> int:
    """
    Base-level worker function to validate schema instances
    """
    try:
        schema_dir = Path(__file__).parent / "schemas"
        schema_path = schema_dir / schema_file_name
        schema = loads(schema_path.read_text())
        validate(instance=instance, schema=schema)
        return 0
    except ValidationError:
        print(error_msg, file=sys.stderr)
        return 1


def validate_file_structure(instance: dict) -> int:
    return validate_schema_instance(
        schema_file_name="file_structure.schema.json",
        instance=instance,
        error_msg="Errors in input file structure. See demo files for examples."
    )


def validate_fluid(instance: dict) -> int:
    fluid_name = str(instance["fluid_name"]).upper()
    instance["fluid_name"] = fluid_name
    return validate_schema_instance(
        schema_file_name="fluid.schema.json",
        instance=instance,
        error_msg="Errors in \"fluid\" input object."
    )


def validate_grout(instance: dict) -> int:
    return validate_schema_instance(
        schema_file_name="grout.schema.json",
        instance=instance,
        error_msg="Errors in \"grout\" input object."
    )


def validate_soil(instance: dict) -> int:
    return validate_schema_instance(
        schema_file_name="soil.schema.json",
        instance=instance,
        error_msg="Errors in \"soil\" input object."
    )


def validate_pipe(instance: dict) -> int:
    pipe_arrangement = str(instance["arrangement"]).upper()
    instance["arrangement"] = pipe_arrangement

    schema_map = {
        BHPipeType.SINGLEUTUBE.name: "pipe_single_double_u_tube.schema.json",
        BHPipeType.DOUBLEUTUBESERIES.name: "pipe_single_double_u_tube.schema.json",
        BHPipeType.DOUBLEUTUBEPARALLEL.name: "pipe_single_double_u_tube.schema.json",
        BHPipeType.COAXIAL.name: "pipe_coaxial.schema.json"
    }

    if pipe_arrangement not in schema_map.keys():
        print("Pipe arrangement not found.", file=sys.stderr)
        return 1

    return validate_schema_instance(
        schema_file_name=schema_map[pipe_arrangement],
        instance=instance,
        error_msg="Errors in \"pipe\" input object."
    )


def validate_borehole(instance: dict) -> int:
    return validate_schema_instance(
        schema_file_name="borehole.schema.json",
        instance=instance,
        error_msg="Errors in \"borehole\" input object."
    )


def validate_simulation(instance: dict) -> int:
    if "timestep" in instance:
        timestep = str(instance["timestep"]).upper()
        instance["timestep"] = timestep

    return validate_schema_instance(
        schema_file_name="simulation.schema.json",
        instance=instance,
        error_msg="Errors in \"simulation\" input object."
    )


def validate_geometric(instance: dict) -> int:
    method = str(instance["method"]).upper()
    instance["method"] = method

    schema_map = {
        DesignGeomType.BIRECTANGLE.name: "geometric_bi_rectangle.schema.json",
        DesignGeomType.BIRECTANGLECONSTRAINED.name: "geometric_bi_rectangle_constrained.schema.json",
        DesignGeomType.BIZONEDRECTANGLE.name: "geometric_bi_zoned_rectangle.schema.json",
        DesignGeomType.NEARSQUARE.name: "geometric_near_square.schema.json",
        DesignGeomType.RECTANGLE.name: "geometric_rectangle.schema.json",
        DesignGeomType.ROWWISE.name: "geometric_rowwise.schema.json",
    }

    if method not in schema_map.keys():
        print("Geometric constraint method not recognized.", file=sys.stderr)
        return 1

    return validate_schema_instance(
        schema_file_name=schema_map[method],
        instance=instance,
        error_msg="Errors in \"geometric_constraints\" input object."
    )


def validate_design(instance: dict) -> int:
    flow_type = str(instance["flow_type"]).upper()
    instance["flow_type"] = flow_type
    return validate_schema_instance(
        schema_file_name="design.schema.json",
        instance=instance,
        error_msg="Errors in \"design\" input object."
    )


def validate_input_file(input_file_path: Path) -> int:
    """
    Validate input file against all schemas
    """

    # get instance data
    instance = loads(input_file_path.read_text())

    # validate
    err_count = 0
    err_count += validate_file_structure(instance)
    err_count += validate_fluid(instance["fluid"])
    err_count += validate_grout(instance["grout"])
    err_count += validate_soil(instance["soil"])
    err_count += validate_pipe(instance["pipe"])
    err_count += validate_borehole(instance["borehole"])
    err_count += validate_simulation(instance["simulation"])
    err_count += validate_geometric(instance["geometric_constraints"])
    err_count += validate_design(instance["design"])
    return err_count


if __name__ == "__main__":
    instance_path = Path(sys.argv[1]).resolve()
    sys.exit(validate_input_file(instance_path))



import csv
import os
import re
import warnings
from datetime import datetime
from json import dumps
from math import floor
from pathlib import Path

from ghedesigner.borehole_heat_exchangers import CoaxialPipe, GHEDesignerBoreholeBase
from ghedesigner.design import AnyBisectionType
from ghedesigner.enums import TimestepType


class OutputManager:

    def __init__(self,
                 design: AnyBisectionType,
                 time: float,
                 project_name: str,
                 notes: str,
                 author: str,
                 model_name: str,
                 load_method: TimestepType,
                 allocated_width=100,
                 ):

        # this constructor should take all the args to build out a full output manager
        # then the client code can decide what to do -- just access data through functions?
        # write all the data to files in a directory?
        # make individual hidden worker functions to build out each part
        # but then add individual public functions to write specific files
        # have one routine to write all of them
        self.text_summary = self.get_summary_text(allocated_width, project_name, model_name, notes, author, time,
                                                  design, load_method)
        self.loading_data_rows = self.get_loading_data(design)
        self.borehole_location_data_rows = self.get_borehole_location_data(design)
        self.hourly_loading_data_rows = self.get_hourly_loading_data(design)
        self.g_function_data_rows = self.get_g_function_data(design)
        self.output_dict = self.get_summary_object(design, time, project_name, notes, author, model_name, load_method)

    def write_all_output_files(self, output_directory: Path, file_suffix: str = ""):
        output_directory.mkdir(exist_ok=True)
        (output_directory / f"SimulationSummary{file_suffix}.txt").write_text(self.text_summary)
        with open(os.path.join(output_directory, f"TimeDependentValues{file_suffix}.csv"), "w", newline="") as csv1OF:
            csv.writer(csv1OF).writerows(self.loading_data_rows)
        with open(os.path.join(output_directory, f"BoreFieldData{file_suffix}.csv"), "w", newline="") as f_csv:
            csv.writer(f_csv).writerows(self.borehole_location_data_rows)
        with open(os.path.join(output_directory, f"Loadings{file_suffix}.csv"), "w", newline="") as f_csv:
            csv.writer(f_csv).writerows(self.hourly_loading_data_rows)
        with open(os.path.join(output_directory, f"Gfunction{file_suffix}.csv"), "w", newline="") as f_csv:
            csv.writer(f_csv).writerows(self.g_function_data_rows)
        with open(str(output_directory / f"SimulationSummary{file_suffix}.json"), "w", newline="") as f_json:
            f_json.write(dumps(self.output_dict, indent=2))

    def get_loading_data(self, design):
        csv_array = [
            [
                "Time (hr)",
                "Time (month)",
                "Q (Rejection) (w) (before time)",
                "Q (Rejection) (W/m) (before time)",
                "Tb (C)",
                "GHE ExFT (C)",
            ]
        ]
        loading_values = design.ghe.loading
        for i, (tv, d_tb, lv) in enumerate(zip(design.ghe.times, design.ghe.dTb, loading_values)):
            if i + 1 < len(design.ghe.times):
                current_time = tv
                loading = loading_values[i + 1]
                current_month = self.hours_to_month(tv)
                normalized_loading = loading / (design.ghe.bhe.b.H * design.ghe.nbh)
                wall_temperature = design.ghe.bhe.soil.ugt + d_tb
                hp_eft_val = design.ghe.hp_eft[i]
                csv_row = list()
                csv_row.append(tv)
                csv_row.append(self.hours_to_month(tv))
                if i > 1:
                    csv_row.append(lv)
                    csv_row.append(lv / (design.ghe.bhe.b.H * design.ghe.nbh))
                else:
                    csv_row.append(0)
                    csv_row.append(0)
                csv_row.append(design.ghe.bhe.soil.ugt + design.ghe.dTb[i - 1])
                csv_row.append(design.ghe.hp_eft[i - 1])
                csv_array.append(csv_row)
            else:
                csv_row = list()
                csv_row.append(tv)
                csv_row.append(self.hours_to_month(tv))
                if i > 1:
                    csv_row.append(lv)
                    csv_row.append(lv / (design.ghe.bhe.b.H * design.ghe.nbh))
                else:
                    csv_row.append(0)
                    csv_row.append(0)
                csv_row.append(design.ghe.bhe.soil.ugt + design.ghe.dTb[i - 1])
                csv_row.append(design.ghe.hp_eft[i - 1])
                csv_array.append(csv_row)

                current_time = tv
                loading = 0
                current_month = self.hours_to_month(tv)
                normalized_loading = loading / (design.ghe.bhe.b.H * design.ghe.nbh)
                wall_temperature = design.ghe.bhe.soil.ugt + d_tb
                hp_eft_val = design.ghe.hp_eft[i]
            csv_row = list()
            csv_row.append(current_time)
            csv_row.append(current_month)
            csv_row.append(loading)
            csv_row.append(normalized_loading)
            csv_row.append(wall_temperature)
            csv_row.append(hp_eft_val)
            csv_array.append(csv_row)
        return csv_array

    @staticmethod
    def get_borehole_location_data(design):
        csv_array = list()
        csv_array.append(["x", "y"])
        for bL in design.ghe.gFunction.bore_locations:
            csv_array.append([bL[0], bL[1]])
        return csv_array

    def get_hourly_loading_data(self, design):
        hourly_loadings = design.ghe.hourly_extraction_ground_loads
        csv_array = list()
        csv_array.append(
            ["Month", "Day", "Hour", "Time (Hours)", "Loading (W) (Extraction)"]
        )
        for hour, hour_load in enumerate(hourly_loadings):
            month, day_in_month, hour_in_day = self.ghe_time_convert(hour)
            csv_array.append([month, day_in_month, hour_in_day, hour, hour_load])
        return csv_array

    @staticmethod
    def get_g_function_data(design):
        title = f"H:{design.ghe.bhe.b.H:0.2f}"
        csv_array = [["ln(t/ts)", f"{title}", f"{title}_bhw"]]
        gf_adjusted, gf_bhw_adjusted = design.ghe.grab_g_function(design.ghe.B_spacing / float(design.ghe.bhe.b.H))

        gf_log_vals = gf_adjusted.x
        gf_g_vals = gf_adjusted.y

        gf_bhw_g_vals = gf_bhw_adjusted.y

        for log_val, g_val, g_bhw_val in zip(gf_log_vals, gf_g_vals, gf_bhw_g_vals):
            csv_array.append([log_val, g_val, g_bhw_val])
        return csv_array

    @staticmethod
    def get_timestep_str(load_method: TimestepType):
        if load_method == TimestepType.HYBRID:
            return TimestepType.HYBRID.name
        if load_method == TimestepType.HOURLY:
            return TimestepType.HOURLY.name
        warnings.warn("Load method not implemented")
        return ""

    def get_summary_object(self,
                           design: AnyBisectionType,
                           time: float,
                           project_name: str,
                           notes: str,
                           author: str,
                           model_name: str,
                           load_method: TimestepType) -> dict:
        # gFunction LTS Table
        g_function_col_titles = ["ln(t/ts)"]
        for g_function_name in list(design.ghe.gFunction.g_lts):
            g_function_col_titles.append("H:" + str(round(g_function_name, 0)) + "m")
        g_function_col_titles.append("H:" + str(round(design.ghe.bhe.b.H, 2)) + "m")
        g_function_data = []
        ghe_gf = design.ghe.gFunction.g_function_interpolation(float(design.ghe.B_spacing) / design.ghe.bhe.b.H)[0]
        for i in range(len(design.ghe.gFunction.log_time)):
            gf_row = list()
            gf_row.append(design.ghe.gFunction.log_time[i])
            for g_function_name in list(design.ghe.gFunction.g_lts):
                gf_row.append(design.ghe.gFunction.g_lts[g_function_name][i])
            gf_row.append(ghe_gf[i])
            g_function_data.append(gf_row)

        def add_with_units(val, units):
            return {'units': units, 'value': val}

        # these are dependent on the # pipes in each borehole, so precalculate
        if isinstance(design.ghe.bhe.pipe.r_out, float):
            pipe_geometry = {'pipe_outer_diameter': add_with_units(design.ghe.bhe.pipe.r_out * 2.0, 'm'),
                             'pipe_inner_diameter': add_with_units(design.ghe.bhe.pipe.r_in * 2.0, 'm')}
            reynolds = GHEDesignerBoreholeBase.compute_reynolds(design.ghe.bhe.m_flow_borehole,
                                                                design.ghe.bhe.pipe.r_in,
                                                                design.ghe.bhe.fluid)
        else:
            pipe_geometry = {
                'inner_pipe_inner_diameter': add_with_units(design.ghe.bhe.pipe.r_in[0] * 2.0, 'm'),
                'inner_pipe_outer_diameter': add_with_units(design.ghe.bhe.pipe.r_in[1] * 2.0, 'm'),
                'outer_pipe_inner_diameter': add_with_units(design.ghe.bhe.pipe.r_out[0] * 2.0, 'm'),
                'outer_pipe_outer_diameter': add_with_units(design.ghe.bhe.pipe.r_out[1] * 2.0, 'm'),
            }
            reynolds = CoaxialPipe.compute_reynolds_concentric(design.ghe.bhe.m_flow_borehole,
                                                               design.ghe.bhe.r_in_out,
                                                               design.ghe.bhe.r_out_in,
                                                               design.ghe.bhe.fluid)
        # build out the actual output dictionary
        output_dict = {
            'project_name': project_name,
            'notes': notes,
            'model_name': model_name,
            'simulation_time_stamp': datetime.now().strftime("%m/%d/%Y %H:%M:%S %p"),
            'simulation_author': author,
            'simulation_runtime': add_with_units(time, 's'),
            'design_selection_search_log': {
                'titles': ["Field", "Excess Temperature", "Max Temperature", "Min Temperature"],
                'units': [" ", "(C)", "(C)", "(C)"],
                'data': design.searchTracker
            },
            'ghe_system': {
                'search_log': {
                    'titles': g_function_col_titles,
                    'units': None,
                    'data': g_function_data
                },
                'active_borehole_length': add_with_units(design.ghe.bhe.b.H, 'm'),
                'borehole_diameter': add_with_units(design.ghe.bhe.b.r_b * 2.0, 'm'),
                'borehole_buried_depth': add_with_units(design.ghe.bhe.b.D, 'm'),
                'borehole_spacing': add_with_units(design.ghe.B_spacing, 'm'),
                'total_drilling': add_with_units(design.ghe.bhe.b.H * len(design.ghe.gFunction.bore_locations), 'm'),
                'field_type': design.ghe.fieldType,
                'field_specifier': design.ghe.fieldSpecifier,
                'number_of_boreholes': len(design.ghe.gFunction.bore_locations),
                'shank_spacing': add_with_units(design.ghe.bhe.pipe.s, 'm'),
                'pipe_geometry': pipe_geometry,
                'pipe_roughness': add_with_units(design.ghe.bhe.pipe.roughness, 'm'),
                'pipe_thermal_conductivity': add_with_units(design.ghe.bhe.pipe.k, 'W/m-K'),
                'pipe_volumetric_heat_capacity': add_with_units(design.ghe.bhe.pipe.rhoCp / 1000, 'kJ/m3-K'),
                'grout_thermal_conductivity': add_with_units(design.ghe.bhe.grout.k, 'W/m-K', ),
                'grout_volumetric_heat_capacity': add_with_units(design.ghe.bhe.grout.rhoCp / 1000, 'kJ/m3-K'),
                'reynolds_number': reynolds,
                'effective_borehole_resistance': add_with_units(design.ghe.bhe.calc_effective_borehole_resistance(),
                                                                'W/m-K'),
                # TODO: are the units right here?
                'soil_thermal_conductivity': add_with_units(design.ghe.bhe.soil.k, 'W/m-K'),
                'soil_volumetric_heat_capacity': add_with_units(design.ghe.bhe.soil.rhoCp / 1000, 'kJ/m3-K'),
                'soil_undisturbed_ground_temp': add_with_units(design.ghe.bhe.soil.ugt, 'C'),
                'fluid_volumetric_heat_capacity': add_with_units(design.ghe.bhe.fluid.rhoCp / 1000, 'kJ/m3-K'),
                'fluid_thermal_conductivity': add_with_units(design.ghe.bhe.fluid.k, 'W/m-K'),
                'fluid_viscosity': add_with_units(design.ghe.bhe.fluid.dynamic_viscosity(), 'Pa-s'),
                'fluid_mixture': design.ghe.bhe.fluid.fluid.fluid_name,  # TODO: Is this the right lookup!?!?!? :)
                'fluid_density': add_with_units(design.ghe.bhe.fluid.rho, 'kg/m3'),
                'fluid_mass_flow_rate_per_borehole': add_with_units(design.ghe.bhe.m_flow_borehole, 'kg/s'),
            },
            'simulation_parameters': {
                'start_month': design.ghe.sim_params.start_month,
                'end_month': design.ghe.sim_params.end_month,
                'maximum_allowable_hp_eft': add_with_units(design.ghe.sim_params.max_EFT_allowable, 'C'),
                'minimum_allowable_hp_eft': add_with_units(design.ghe.sim_params.min_EFT_allowable, 'C'),
                'maximum_allowable_height': add_with_units(design.ghe.sim_params.max_height, 'm'),
                'minimum_allowable_height': add_with_units(design.ghe.sim_params.min_height, 'm'),
                'simulation_time': add_with_units(int(design.ghe.sim_params.end_month / 12), 'years'),
                'simulation_load_method': self.get_timestep_str(load_method)
            },
            'simulation_results': {

            }

        }

        # potentially add convection coefficient -- not sure why we wouldn't do it
        if hasattr(design.ghe.bhe, "h_f"):
            # TODO: Should be W/m2-K?
            output_dict['ghe_system']['fluid_convection_coefficient'] = add_with_units(design.ghe.bhe.h_f, 'W/m-K')

        # add monthly load summary
        monthly_load_values = []
        n_months = len(design.ghe.hybrid_load.monthly_cl) - 1
        n_years = int(n_months / 12)
        months = n_years * [
            "January", "February", "March", "April", "May", "June",
            "July", "August", "September", "October", "November", "December",
        ]
        start_ind = 1
        stop_ind = n_months
        for i in range(start_ind, stop_ind + 1):
            monthly_load_values.append(
                [
                    months[i - 1],
                    design.ghe.hybrid_load.monthly_hl[i],
                    design.ghe.hybrid_load.monthly_cl[i],
                    design.ghe.hybrid_load.monthly_peak_hl[i],
                    design.ghe.hybrid_load.monthly_peak_hl_duration[i],
                    design.ghe.hybrid_load.monthly_peak_cl[i],
                    design.ghe.hybrid_load.monthly_peak_cl_duration[i]]
            )
        output_dict['ghe_system']['glhe_monthly_loads'] = {
            'titles': [
                "Month",
                "Total Heating",
                "Total Cooling",
                "Peak Heating",
                "PH Duration",
                "Peak Cooling",
                "PC Duration",
            ],
            'units': ["", "kWh", "kWh", "kW", "hr", "kW", "hr"],
            'data': monthly_load_values
        }

        # add simulation results stuff
        n_years = 0
        out_array = []
        last_month = -1
        month_tb_vals = []
        month_eft_vals = []
        for tv, d_tb, eft in zip(design.ghe.times, design.ghe.dTb, design.ghe.hp_eft):
            current_month = floor(self.hours_to_month(tv))
            if current_month == last_month:
                month_tb_vals.append(d_tb)
                month_eft_vals.append(eft)
            elif current_month != last_month:
                if len(month_tb_vals) > 0:
                    if len(out_array) == 0:
                        previous_temp = design.ghe.bhe.soil.ugt
                    else:
                        previous_temp = design.ghe.bhe.soil.ugt
                    out_array.append(
                        [
                            current_month,
                            previous_temp + month_tb_vals[-1],
                            max(month_eft_vals),
                            min(month_eft_vals),
                        ]
                    )
                last_month = current_month
                month_tb_vals = [d_tb]
                month_eft_vals = [eft]
            if current_month % 11 == 0:
                n_years += 1
        max_eft = max(design.ghe.hp_eft)
        min_eft = min(design.ghe.hp_eft)
        max_eft_time = design.ghe.times[design.ghe.hp_eft.index(max(design.ghe.hp_eft))]
        min_eft_time = design.ghe.times[design.ghe.hp_eft.index(min(design.ghe.hp_eft))]
        max_eft_time = self.hours_to_month(max_eft_time)
        min_eft_time = self.hours_to_month(min_eft_time)
        output_dict['simulation_results'] = {
            'max_hp_eft': add_with_units(max_eft, 'C'),
            'max_hp_eft_time': add_with_units(max_eft_time, 'months'),
            'min_hp_eft': add_with_units(min_eft, 'C'),
            'min_hp_eft_time': add_with_units(min_eft_time, 'months'),
            'monthly_temp_summary': {
                'titles': ["Time", "Tbw", "Max hp_eft", "Min hp_eft"],
                'units': ["(months)", "(C)", "(C)", "(C)"],
                'data': out_array
            }
        }

        return output_dict

    def get_summary_text(self, width, project_name, model_name, notes, author, time, design, load_method):

        f_int = ".0f"
        f_1f = ".1f"
        f_2f = ".2f"
        f_3f = ".3f"
        f_4f = ".4f"
        f_str = "s"
        f_sci = ".3e"

        blank_line = self.create_line(width)
        empty_line = self.create_line(width, character=" ")

        o = blank_line
        o += self.d_row(width, "Project Name:", project_name, f_str)
        o += blank_line
        o += "Notes:\n\n" + notes + "\n"
        o += blank_line
        o += self.d_row(width, "File/Model Name:", model_name, f_str)
        now = datetime.now()
        time_string = now.strftime("%m/%d/%Y %H:%M:%S %p")
        o += self.d_row(width, "Simulated On:", time_string, f_str)
        o += self.d_row(width, "Simulated By:", author, f_str)
        o += self.d_row(width, "Calculation Time, s:", time, f_3f)
        o += empty_line
        o += self.create_title(width, "Design Selection", filler_symbol="-")

        design_header = [
            ["Field", "Excess Temperature", "Max Temperature", "Min Temperature"],
            [" ", "(C)", "(C)", "(C)"],
        ]
        try:
            design_values = design.searchTracker
        except:
            design_values = ""
        design_formats = [f_str, f_2f, f_2f, f_2f]

        o += self.create_table("Field Search Log", design_header, design_values, width, design_formats,
                               filler_symbol="-", centering="^")

        o += empty_line
        o += self.create_title(width, "GHE System", filler_symbol="-")

        # gFunction LTS Table
        g_function_table_formats = [f_3f]
        gf_table_ff = [f_3f] * (len(design.ghe.gFunction.g_lts) + 1)
        g_function_table_formats.extend(gf_table_ff)
        g_function_col_titles = ["ln(t/ts)"]

        for g_function_name in list(design.ghe.gFunction.g_lts):
            g_function_col_titles.append("H:" + str(round(g_function_name, 0)) + "m")
        g_function_col_titles.append("H:" + str(round(design.ghe.bhe.b.H, 2)) + "m")

        g_function_data = []
        ghe_gf = design.ghe.gFunction.g_function_interpolation(float(design.ghe.B_spacing) / design.ghe.bhe.b.H)[0]
        for i in range(len(design.ghe.gFunction.log_time)):
            gf_row = list()
            gf_row.append(design.ghe.gFunction.log_time[i])
            for g_function_name in list(design.ghe.gFunction.g_lts):
                gf_row.append(design.ghe.gFunction.g_lts[g_function_name][i])
            gf_row.append(ghe_gf[i])
            g_function_data.append(gf_row)

        o += self.create_table("gFunction LTS Values", [g_function_col_titles], g_function_data, width,
                               g_function_table_formats, filler_symbol="-", centering="^")
        o += empty_line

        o += self.create_title(width, "System Parameters", filler_symbol="-")
        o += self.d_row(width, "Active Borehole Length, m:", design.ghe.bhe.b.H, f_int)
        o += self.d_row(width, "Borehole Diameter, mm:", design.ghe.bhe.b.r_b * 1000 * 2.0, f_2f)
        o += self.d_row(width, "Borehole Spacing, m:", design.ghe.B_spacing, f_3f)
        o += self.d_row(width, 'Borhole Depth, m:', design.ghe.bhe.b.D, f_2f)
        o += self.d_row(width, "Total Drilling, m:", design.ghe.bhe.b.H * len(design.ghe.gFunction.bore_locations),
                        f_int)

        o += "Field Geometry: " + "\n"
        o += self.d_row(width, "Field Type:", design.ghe.fieldType, f_str, n_tabs=1)
        o += self.d_row(width, "Field Specifier:", design.ghe.fieldSpecifier, f_str, n_tabs=1)
        o += self.d_row(width, "NBH:", len(design.ghe.gFunction.bore_locations), f_int, n_tabs=1)

        o += "Borehole Information: " + "\n"

        if isinstance(design.ghe.bhe.pipe.r_out, float):
            o += self.d_row(width, "Pipe Outer Diameter, mm:", design.ghe.bhe.pipe.r_out * 1000 * 2.0, f_2f, n_tabs=1)
            o += self.d_row(width, "Pipe Inner Diameter, mm:", design.ghe.bhe.pipe.r_in * 1000 * 2.0, f_2f, n_tabs=1)
        else:
            o += self.d_row(width, "Outer Pipe Outer Diameter, mm:", design.ghe.bhe.pipe.r_out[1] * 1000 * 2.0, f_2f, n_tabs=1)
            o += self.d_row(width, "Outer Pipe Inner Diameter, mm:", design.ghe.bhe.pipe.r_out[0] * 1000 * 2.0, f_2f, n_tabs=1)
            o += self.d_row(width, "Inner Pipe Outer Diameter, mm:", design.ghe.bhe.pipe.r_in[1] * 1000 * 2.0, f_2f, n_tabs=1)
            o += self.d_row(width, "Inner Pipe Inner Diameter, mm:", design.ghe.bhe.pipe.r_in[0] * 1000 * 2.0, f_2f, n_tabs=1)

        o += self.d_row(width, "Pipe Roughness, m:", design.ghe.bhe.pipe.roughness, f_sci, n_tabs=1)
        if isinstance(design.ghe.bhe.pipe.k, float):
            o += self.d_row(width, "Pipe Thermal Conductivity, W/m-K:", design.ghe.bhe.pipe.k, f_3f, n_tabs=1)
        else:
            o += self.d_row(width, "Inner Pipe Thermal Conductivity, W/m-K:", design.ghe.bhe.pipe.k[0], f_3f, n_tabs=1)
            o += self.d_row(width, "Outer Pipe Thermal Conductivity, W/m-K:", design.ghe.bhe.pipe.k[1], f_3f, n_tabs=1)

        o += self.d_row(width, "Pipe Volumetric Heat Capacity, kJ/m3-K:", design.ghe.bhe.pipe.rhoCp / 1000, f_2f, n_tabs=1)
        o += self.d_row(width, "Shank Spacing, mm:", design.ghe.bhe.pipe.s * 1000, f_2f, n_tabs=1)
        o += self.d_row(width, "Grout Thermal Conductivity, W/(m-K):", design.ghe.bhe.grout.k, f_3f, n_tabs=1)
        o += self.d_row(width, "Grout Volumetric Heat Capacity, kJ/m3-K:", design.ghe.bhe.grout.rhoCp / 1000, f_2f,
                        n_tabs=1)
        if isinstance(design.ghe.bhe.pipe.r_out, float):
            o += self.d_row(width, "Reynold's Number:",
                            GHEDesignerBoreholeBase.compute_reynolds(design.ghe.bhe.m_flow_borehole,
                                                                     design.ghe.bhe.pipe.r_in,
                                                                     design.ghe.bhe.fluid),
                            f_int, n_tabs=1)
        else:

            o += self.d_row(width, "Reynold's Number:",
                            CoaxialPipe.compute_reynolds_concentric(design.ghe.bhe.m_flow_borehole,
                                                                    design.ghe.bhe.r_in_out,
                                                                    design.ghe.bhe.r_out_in,
                                                                    design.ghe.bhe.fluid),
                            f_int, n_tabs=1)

        o += self.d_row(width, "Effective Borehole Resistance, W/m-K:",
                        design.ghe.bhe.calc_effective_borehole_resistance(),
                        f_4f, n_tabs=1)
        # Shank Spacing, Pipe Type, etc.

        o += "Soil Properties: " + "\n"
        o += self.d_row(width, "Thermal Conductivity, W/m-K:", design.ghe.bhe.soil.k, f_3f, n_tabs=1)
        o += self.d_row(width, "Volumetric Heat Capacity, kJ/m3-K:",
                        design.ghe.bhe.soil.rhoCp / 1000, f_2f, n_tabs=1)
        o += self.d_row(width, "Undisturbed Ground Temperature, C:", design.ghe.bhe.soil.ugt, f_2f, n_tabs=1)

        o += "Fluid Properties" + "\n"
        o += self.d_row(width, "Volumetric Heat Capacity, kJ/m3-K:",
                        design.ghe.bhe.fluid.rhoCp / 1000, f_2f, n_tabs=1)
        o += self.d_row(width, "Thermal Conductivity, W/m-K:", design.ghe.bhe.fluid.k, f_2f, n_tabs=1)
        o += self.d_row(width, "Viscosity, Pa-s:", design.ghe.bhe.fluid.dynamic_viscosity(), f_sci, n_tabs=1)
        o += self.d_row(width, "Fluid Mix:", design.ghe.bhe.fluid.fluid.fluid_name, f_str, n_tabs=1)
        o += self.d_row(width, "Density, kg/m^3:", design.ghe.bhe.fluid.rho, f_2f, n_tabs=1)
        o += self.d_row(width, "Mass Flow Rate Per Borehole, kg/s:", design.ghe.bhe.m_flow_borehole, f_3f, n_tabs=1)
        if hasattr(design.ghe.bhe, "h_f"):
            o += self.d_row(width, "Fluid Convection Coefficient, W/m-K:", design.ghe.bhe.h_f, f_int, n_tabs=1)
        o += empty_line

        monthly_load_values = []
        n_months = len(design.ghe.hybrid_load.monthly_cl) - 1
        n_years = int(n_months / 12)
        months = n_years * [
            "January", "February", "March", "April", "May", "June",
            "July", "August", "September", "October", "November", "December",
        ]

        start_ind = 1
        stop_ind = n_months
        for i in range(start_ind, stop_ind + 1):
            monthly_load_values.append(
                [
                    months[i - 1],
                    design.ghe.hybrid_load.monthly_hl[i],
                    design.ghe.hybrid_load.monthly_cl[i],
                    design.ghe.hybrid_load.monthly_peak_hl[i],
                    design.ghe.hybrid_load.monthly_peak_hl_duration[i],
                    design.ghe.hybrid_load.monthly_peak_cl[i],
                    design.ghe.hybrid_load.monthly_peak_cl_duration[i]
                ]
            )
        month_header = [
            ["Month", "Total Heating", "Total Cooling", "Peak Heating", "PH Duration", "Peak Cooling", "PC Duration"],
            ["", "kWh", "kWh", "kW", "hr", "kW", "hr"],
        ]

        month_table_formats = [f_str, f_1f, f_1f, f_1f, f_1f, f_1f, f_1f]

        o += self.create_table("GLHE Monthly Loads", month_header, monthly_load_values, width, month_table_formats,
                               filler_symbol="-", centering="^")

        o += empty_line

        o += self.create_title(width, "Simulation Parameters")
        o += self.d_row(width, "Start Month: ", design.ghe.sim_params.start_month, f_int)
        o += self.d_row(width, "End Month: ", design.ghe.sim_params.end_month, f_int)
        o += self.d_row(width, "Maximum Allowable hp_eft, C: ", design.ghe.sim_params.max_EFT_allowable, f_2f)
        o += self.d_row(width, "Minimum Allowable hp_eft, C: ", design.ghe.sim_params.min_EFT_allowable, f_2f)
        o += self.d_row(width, "Maximum Allowable Height, m: ", design.ghe.sim_params.max_height, f_2f)
        o += self.d_row(width, "Minimum Allowable Height, m: ", design.ghe.sim_params.min_height, f_2f)
        o += self.d_row(width, "Simulation Time, years: ", int(design.ghe.sim_params.end_month / 12), f_int)
        load_method_string = self.get_timestep_str(load_method)
        o += self.d_row(width, "Simulation Loading Type: ", load_method_string, f_str)

        o += empty_line

        # Loading Stuff
        o += self.create_title(width, "Simulation Results")
        o += empty_line

        # Simulation Results
        eft_table_title = "Monthly Temperature Summary"
        n_years = 0
        out_array = []
        last_month = -1
        month_tb_vals = []
        month_eft_vals = []
        for tv, d_tb, eft in zip(design.ghe.times, design.ghe.dTb, design.ghe.hp_eft):
            # currentHourMonth = timeVals[i] - hTotalYear * nYears
            current_month = floor(self.hours_to_month(tv))
            # print(monthEFTVals)
            if current_month == last_month:
                month_tb_vals.append(d_tb)
                month_eft_vals.append(eft)
            elif current_month != last_month:
                if len(month_tb_vals) > 0:
                    if len(out_array) == 0:
                        previous_temp = design.ghe.bhe.soil.ugt
                    else:
                        # print(i)
                        # print(len(outArray))
                        # previousTemp = outArray[-1][1]
                        previous_temp = design.ghe.bhe.soil.ugt
                    out_array.append(
                        [
                            current_month,
                            previous_temp + month_tb_vals[-1],
                            max(month_eft_vals),
                            min(month_eft_vals),
                        ]
                    )
                last_month = current_month
                month_tb_vals = [d_tb]
                month_eft_vals = [eft]
            if current_month % 11 == 0:
                n_years += 1

        header_array = [
            ["Time", "Tbw", "Max hp_eft", "Min hp_eft"],
            ["(months)", "(C)", "(C)", "(C)"],
        ]
        eft_table_formats = [f_int, f_2f, f_2f, f_2f]

        o += self.create_title(width, "Peak Temperature", filler_symbol="-")
        max_eft = max(design.ghe.hp_eft)
        min_eft = min(design.ghe.hp_eft)
        max_eft_time = design.ghe.times[design.ghe.hp_eft.index(max(design.ghe.hp_eft))]
        min_eft_time = design.ghe.times[design.ghe.hp_eft.index(min(design.ghe.hp_eft))]
        max_eft_time = self.hours_to_month(max_eft_time)
        min_eft_time = self.hours_to_month(min_eft_time)
        o += self.d_row(width, "Max hp_eft, C:", max_eft, f_3f)
        o += self.d_row(width, "Max hp_eft Time, Months:", max_eft_time, f_3f)
        o += self.d_row(width, "Min hp_eft, C:", min_eft, f_3f)
        o += self.d_row(width, "Min hp_eft Time, Months:", min_eft_time, f_3f)

        o += self.create_table(eft_table_title, header_array, out_array, width, eft_table_formats, filler_symbol="-",
                               centering="^")

        # strip out all trailing whitespace
        o = re.sub(r"\s+\n", "\n", o)

        return o

    @staticmethod
    def create_title(allocated_width, title, filler_symbol=" "):
        return "{:{fS}^{L}s}\n".format(" " + title + " ", L=allocated_width, fS=filler_symbol)

    @staticmethod
    def create_row(allocated_width, row_data, data_formats, centering=">"):
        r_s = ""
        n_cols = len(row_data)
        col_width = int(allocated_width / n_cols)
        left_over = float(allocated_width) % col_width
        for d_f, data in zip(data_formats, row_data):
            width = col_width
            if left_over > 0:
                width = col_width + 1
                left_over -= 1
            try:
                r_s += "{:{c}{w}{fm}}".format(data, c=centering, w=width, fm=d_f)
            except:
                print("Output Row creation error: ", d_f)
                raise ValueError

        r_s += "\n"
        return r_s

    @staticmethod
    def create_table(title, col_titles, rows, allocated_width, col_formats, filler_symbol=" ", centering=">"):
        n_cols = len(col_titles[0])
        r_s = ""
        r_s += OutputManager.create_title(allocated_width, title, filler_symbol=filler_symbol)
        blank_line = OutputManager.create_line(allocated_width)
        r_s += blank_line
        header_format = ["s"] * n_cols
        for colT in col_titles:
            r_s += OutputManager.create_row(allocated_width, colT, header_format, centering="^")
        r_s += blank_line
        for row in rows:
            r_s += OutputManager.create_row(allocated_width, row, col_formats, centering=centering)
        r_s += blank_line
        return r_s

    @staticmethod
    def d_row(row_allocation: int, entry_1: str, entry_2, d_type: str, n_tabs: int = 0):
        tab_width = 4
        leading_spaces = n_tabs * tab_width

        l_str = f"{' ' * leading_spaces}{entry_1}"
        r_str = f"{entry_2:{d_type}}"

        l_chars_needed = len(l_str)
        r_chars_needed = len(r_str)
        c_spaces = row_allocation - l_chars_needed - r_chars_needed

        if c_spaces < 0:
            warnings.warn("Unable to write output string with specified formatting.")
            c_spaces = 4

        c_str = ' ' * c_spaces
        r_s = f"{l_str}{c_str}{r_str}\n"
        return r_s

    @staticmethod
    def create_line(row_allocation, character="*"):
        return character * row_allocation + "\n"

    @staticmethod
    def hours_to_month(hours):
        days_in_year = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
        hours_in_year = [24 * x for x in days_in_year]
        n_years = floor(hours / sum(hours_in_year))
        frac_month = n_years * len(days_in_year)
        month_in_year = 0
        for idx, _ in enumerate(days_in_year):
            hours_left = hours - n_years * sum(hours_in_year)
            if sum(hours_in_year[0: idx + 1]) >= hours_left:
                month_in_year = idx
                break
        frac_month += month_in_year
        h_l = hours - n_years * sum(hours_in_year) - sum(hours_in_year[0:month_in_year])
        frac_month += h_l / (hours_in_year[month_in_year])
        return frac_month

    @staticmethod
    def ghe_time_convert(hours):
        days_in_year = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
        hours_in_year = [24 * x for x in days_in_year]
        month_in_year = 0
        year_hour_sum = 0
        for idx, _ in enumerate(days_in_year):
            hours_left = hours
            if year_hour_sum + hours_in_year[idx] - 1 >= hours_left:
                month_in_year = idx
                break
            else:
                year_hour_sum += hours_in_year[idx]
        h_l = hours - sum(hours_in_year[0:month_in_year])
        day_in_month = floor(h_l / 24) + 1
        hour_in_day = h_l % 24 + 1
        return month_in_year + 1, day_in_month, hour_in_day



from enum import auto, IntEnum
from math import exp, log, pi, sqrt

import numpy as np
from scipy.interpolate import interp1d
from scipy.linalg.lapack import dgtsv

from ghedesigner.borehole_heat_exchangers import SingleUTube
from ghedesigner.constants import TWO_PI


class CellProps(IntEnum):
    R_IN = 0
    R_CENTER = auto()
    R_OUT = auto()
    K = auto()
    RHO_CP = auto()
    TEMP = auto()
    VOL = auto()


class RadialNumericalBH(object):
    """
    X. Xu and Jeffrey D. Spitler. 2006. 'Modeling of Vertical Ground Loop Heat
    Exchangers with Variable Convective Resistance and Thermal Mass of the
    Fluid.' in Proceedings of the 10th International Conference on Thermal
    Energy Storage-EcoStock. Pomona, NJ, May 31-June 2.
    """

    def __init__(self, single_u_tube: SingleUTube):
        self.single_u_tube = single_u_tube

        # "The one dimensional model has a fluid core, an equivalent convective
        # resistance layer, a tube layer, a grout layer and is surrounded by the
        # ground."

        # cell numbers
        self.num_fluid_cells = 3
        self.num_conv_cells = 1
        self.num_pipe_cells = 4
        self.num_grout_cells = 27
        self.num_soil_cells = 500

        self.num_cells = self.num_fluid_cells + self.num_conv_cells + self.num_fluid_cells
        self.num_cells += self.num_grout_cells + self.num_soil_cells + 1

        self.bh_wall_idx = self.num_fluid_cells + self.num_conv_cells + self.num_pipe_cells + self.num_grout_cells

        # Geometry and grid procedure

        # far-field radius is set to 10m; the soil region is represented by
        # 500 cells
        far_field_radius = 10  # soil radius (in meters)
        self.r_far_field = far_field_radius - single_u_tube.b.r_b

        # borehole radius is set to the actual radius of the borehole
        self.r_borehole = single_u_tube.b.r_b

        # outer tube radius is set to sqrt(2) * r_p_o, tube region has 4 cells
        self.r_out_tube = sqrt(2) * single_u_tube.pipe.r_out

        # inner tube radius is set to r_out_tube-t_p
        self.thickness_pipe = single_u_tube.pipe.r_out - single_u_tube.pipe.r_in
        self.r_in_tube = self.r_out_tube - self.thickness_pipe

        # r_in_convection is set to r_in_tube - 1/4 * t
        self.r_in_convection = self.r_in_tube - self.thickness_pipe / 4.0

        # r_fluid is set to r_in_convection - 3/4 * t
        self.r_fluid = self.r_in_convection - (3.0 / 4.0 * self.thickness_pipe)

        # Thicknesses of the grid regions
        self.thickness_soil = (self.r_far_field - self.r_borehole) / self.num_soil_cells
        self.thickness_grout = (self.r_borehole - self.r_out_tube) / self.num_grout_cells
        # pipe thickness is equivalent to original tube thickness
        self.thickness_conv = (self.r_in_tube - self.r_in_convection) / self.num_conv_cells
        self.thickness_fluid = (self.r_in_convection - self.r_fluid) / self.num_fluid_cells

        # other
        self.init_temp = 20

        # other
        self.g = np.array([], dtype=np.double)
        self.g_bhw = np.array([], dtype=np.double)
        self.lntts = np.array([], dtype=np.double)
        self.c_0 = TWO_PI * single_u_tube.soil.k
        soil_diffusivity = single_u_tube.k_s / single_u_tube.soil.rhoCp
        self.t_s = single_u_tube.b.H ** 2 / (9 * soil_diffusivity)
        # default is at least 49 hours, or up to -8.6 log time
        self.calc_time_in_sec = max([self.t_s * exp(-8.6), 49.0 * 3600.0])
        self.g_sts = None

    def partial_init(self, single_u_tube: SingleUTube):
        # TODO: unravel how to eliminate this.
        # - It was calling the full class ctor "self.__init__()" which is just plain wrong...
        # - Now we're calling a stripped down version with only the most essential
        #   variables which are required.
        # - This is here partially because equivalent boreholes are generated.
        self.single_u_tube = single_u_tube
        soil_diffusivity = single_u_tube.k_s / single_u_tube.soil.rhoCp
        self.t_s = single_u_tube.b.H ** 2 / (9 * soil_diffusivity)
        self.calc_time_in_sec = max([self.t_s * exp(-8.6), 49.0 * 3600.0])

    def fill_radial_cell(self, radial_cell, resist_p_eq, resist_f_eq, resist_tg_eq):

        num_fluid_cells = self.num_fluid_cells
        num_conv_cells = self.num_conv_cells
        num_pipe_cells = self.num_pipe_cells
        num_grout_cells = self.num_grout_cells
        num_soil_cells = self.num_soil_cells

        cell_summation = 0

        # load fluid cells
        for idx in range(cell_summation, num_fluid_cells + cell_summation):
            center_radius = self.r_fluid + idx * self.thickness_fluid

            if idx == 0:
                inner_radius = center_radius
            else:
                inner_radius = center_radius - self.thickness_fluid / 2.0

            outer_radius = center_radius + self.thickness_fluid / 2.0

            # The equivalent thermal mass of the fluid can be calculated from
            # equation (2)
            # pi (r_in_conv ** 2 - r_f **2) C_eq_f = 2pi r_p_in**2 * C_f
            rho_cp_eq = (2.0
                         * (self.single_u_tube.pipe.r_in ** 2)
                         * self.single_u_tube.fluid.rhoCp
                         ) / ((self.r_in_convection ** 2) - (self.r_fluid ** 2))

            k_eq = rho_cp_eq / self.single_u_tube.fluid.cp

            volume = pi * (outer_radius ** 2 - inner_radius ** 2)
            radial_cell[:, idx] = np.array(
                [
                    inner_radius,
                    center_radius,
                    outer_radius,
                    k_eq,
                    rho_cp_eq,
                    self.init_temp,
                    volume,
                ],
                dtype=np.double,
            )
        cell_summation += num_fluid_cells

        # TODO: verify whether errors are possible here and raise exception if needed
        # assert cell_summation == num_fluid_cells

        # load convection cells
        for idx in range(cell_summation, num_conv_cells + cell_summation):
            j = idx - cell_summation
            inner_radius = self.r_in_convection + j * self.thickness_conv
            center_radius = inner_radius + self.thickness_conv / 2.0
            outer_radius = inner_radius + self.thickness_conv
            k_eq = log(self.r_in_tube / self.r_in_convection) / (TWO_PI * resist_f_eq)
            rho_cp = 1.0
            volume = pi * (outer_radius ** 2 - inner_radius ** 2)
            radial_cell[:, idx] = np.array(
                [
                    inner_radius,
                    center_radius,
                    outer_radius,
                    k_eq,
                    rho_cp,
                    self.init_temp,
                    volume,
                ],
                dtype=np.double,
            )
        cell_summation += num_conv_cells

        # TODO: verify whether errors are possible here and raise exception if needed
        # assert cell_summation == (num_fluid_cells + num_conv_cells)

        # load pipe cells
        for idx in range(cell_summation, num_pipe_cells + cell_summation):
            j = idx - cell_summation
            inner_radius = self.r_in_tube + j * self.thickness_pipe
            center_radius = inner_radius + self.thickness_pipe / 2.0
            outer_radius = inner_radius + self.thickness_pipe
            conductivity = log(self.r_borehole / self.r_in_tube) / (TWO_PI * resist_p_eq)
            rho_cp = self.single_u_tube.pipe.rhoCp
            volume = pi * (outer_radius ** 2 - inner_radius ** 2)
            radial_cell[:, idx] = np.array(
                [
                    inner_radius,
                    center_radius,
                    outer_radius,
                    conductivity,
                    rho_cp,
                    self.init_temp,
                    volume,
                ],
                dtype=np.double,
            )
        cell_summation += num_pipe_cells

        # TODO: verify whether errors are possible here and raise exception if needed
        # assert cell_summation == (num_fluid_cells + num_conv_cells + num_pipe_cells)

        # load grout cells
        for idx in range(cell_summation, num_grout_cells + cell_summation):
            j = idx - cell_summation
            inner_radius = self.r_out_tube + j * self.thickness_grout
            center_radius = inner_radius + self.thickness_grout / 2.0
            outer_radius = inner_radius + self.thickness_grout
            conductivity = log(self.r_borehole / self.r_in_tube) / (TWO_PI * resist_tg_eq)
            rho_cp = self.single_u_tube.grout.rhoCp
            volume = pi * (outer_radius ** 2 - inner_radius ** 2)
            radial_cell[:, idx] = np.array(
                [
                    inner_radius,
                    center_radius,
                    outer_radius,
                    conductivity,
                    rho_cp,
                    self.init_temp,
                    volume,
                ],
                dtype=np.double,
            )
        cell_summation += num_grout_cells

        # TODO: verify whether errors are possible here and raise exception if needed
        # assert cell_summation == (num_fluid_cells + num_conv_cells + num_pipe_cells + num_grout_cells)

        # load soil cells
        for idx in range(cell_summation, num_soil_cells + cell_summation):
            j = idx - cell_summation
            inner_radius = self.r_borehole + j * self.thickness_soil
            center_radius = inner_radius + self.thickness_soil / 2.0
            outer_radius = inner_radius + self.thickness_soil
            conductivity = self.single_u_tube.soil.k
            rho_cp = self.single_u_tube.soil.rhoCp
            volume = pi * (outer_radius ** 2 - inner_radius ** 2)
            radial_cell[:, idx] = np.array(
                [
                    inner_radius,
                    center_radius,
                    outer_radius,
                    conductivity,
                    rho_cp,
                    self.init_temp,
                    volume,
                ],
                dtype=np.double,
            )
        cell_summation += num_soil_cells

    def calc_sts_g_functions(self, single_u_tube, final_time=None) -> tuple:

        self.partial_init(single_u_tube)

        resist_bh_effective = self.single_u_tube.calc_effective_borehole_resistance()

        resist_f_eq = self.single_u_tube.R_f / 2.0
        resist_p_eq = self.single_u_tube.R_p / 2.0
        resist_tg_eq = resist_bh_effective - resist_f_eq

        # Pass radial cell by reference and fill here so that it can be
        # destroyed when this method returns
        radial_cell = np.zeros(shape=(len(CellProps), self.num_cells), dtype=np.double)
        self.fill_radial_cell(radial_cell, resist_p_eq, resist_f_eq, resist_tg_eq)

        if final_time is None:
            final_time = self.calc_time_in_sec

        g = []
        g_bhw = []
        lntts = []

        _dl = np.zeros(self.num_cells - 1)
        _d = np.zeros(self.num_cells)
        _du = np.zeros(self.num_cells - 1)
        _b = np.zeros(self.num_cells)

        heat_flux = 1.0
        init_temp = self.init_temp

        time = 1e-12 - 120
        time_step = 120

        _fe_1 = np.zeros(shape=(self.num_cells - 2), dtype=np.double)
        _fe_2 = np.zeros_like(_fe_1)
        _ae = np.zeros_like(_fe_2)
        _fw_1 = np.zeros_like(_ae)
        _fw_2 = np.zeros_like(_fw_1)
        _aw = np.zeros_like(_fw_2)
        _ad = np.zeros_like(_aw)

        _west_cell = radial_cell[:, 0: self.num_cells - 2]
        _center_cell = radial_cell[:, 1: self.num_cells - 1]
        _east_cell = radial_cell[:, 2: self.num_cells - 0]

        fe_1 = log(radial_cell[CellProps.R_OUT, 0] / radial_cell[CellProps.R_CENTER, 0])
        fe_1 /= (TWO_PI * radial_cell[CellProps.K, 0])

        fe_2 = log(radial_cell[CellProps.R_CENTER, 1] / radial_cell[CellProps.R_IN, 1])
        fe_2 /= (TWO_PI * radial_cell[CellProps.K, 1])

        ae = 1 / (fe_1 + fe_2)
        ad = radial_cell[CellProps.RHO_CP, 0] * radial_cell[CellProps.VOL, 0] / time_step
        _d[0] = -ae / ad - 1
        _du[0] = ae / ad

        def fill_f1(fx_1, cell):
            fx_1[:] = np.log(cell[CellProps.R_OUT, :] / cell[CellProps.R_CENTER, :]) / (TWO_PI * cell[CellProps.K, :])

        def fill_f2(fx_2, cell):
            fx_2[:] = np.log(cell[CellProps.R_CENTER, :] / cell[CellProps.R_IN, :]) / (TWO_PI * cell[CellProps.K, :])

        fill_f1(_fe_1, _center_cell)
        fill_f2(_fe_2, _east_cell)
        _ae[:] = 1.0 / (_fe_1 + _fe_2)

        fill_f1(_fw_1, _west_cell)
        fill_f2(_fw_2, _center_cell)
        _aw[:] = -1.0 / (_fw_1 + _fw_2)

        _ad[:] = (_center_cell[CellProps.RHO_CP, :] * _center_cell[CellProps.VOL, :] / time_step)
        _dl[0: self.num_cells - 2] = -_aw / _ad
        _d[1: self.num_cells - 1] = _aw / _ad - _ae / _ad - 1.0
        _du[1: self.num_cells - 1] = _ae / _ad

        while True:

            time += time_step

            # For the idx == 0 case:

            _b[0] = -radial_cell[CellProps.TEMP, 0] - heat_flux / ad

            # For the idx == n-1 case

            _dl[self.num_cells - 2] = 0.0
            _d[self.num_cells - 1] = 1.0
            _b[self.num_cells - 1] = radial_cell[CellProps.TEMP, self.num_cells - 1]

            # Now handle the 1 to n-2 cases with numpy slicing and vectorization
            _b[1: self.num_cells - 1] = -radial_cell[CellProps.TEMP, 1: self.num_cells - 1]

            # Tri-diagonal matrix solver
            # High level interface to LAPACK routine
            # https://docs.scipy.org/doc/scipy/reference/generated/scipy.linalg.lapack.dgtsv.html#scipy.linalg.lapack.dgtsv
            dgtsv(_dl, _d, _du, _b, overwrite_b=1)  # TODO: Do we really need lapack just to do a TDMA solution?

            radial_cell[CellProps.TEMP, :] = _b

            # compute standard g-functions
            g.append(self.c_0 * ((radial_cell[CellProps.TEMP, 0] - init_temp) / heat_flux - resist_bh_effective))

            # compute g-functions at bh wall
            bh_wall_temp = radial_cell[CellProps.TEMP, self.bh_wall_idx]
            g_bhw.append(self.c_0 * ((bh_wall_temp - init_temp) / heat_flux))

            lntts.append(log(time / self.t_s))

            if time >= final_time - time_step:
                break

        # quickly chop down the total values to a more manageable set
        num_intervals = 30
        g_tmp = interp1d(lntts, g)
        uniform_lntts_vals = np.linspace(lntts[0], lntts[-1], num_intervals)
        uniform_g_vals = g_tmp(uniform_lntts_vals)

        g_bhw_tmp = interp1d(lntts, g_bhw)
        uniform_g_bhw_vals = g_bhw_tmp(uniform_lntts_vals)

        # set the final arrays and interpolator objects
        self.lntts = np.array(uniform_lntts_vals)
        self.g = np.array(uniform_g_vals)
        self.g_bhw = np.array(uniform_g_bhw_vals)
        self.g_sts = interp1d(self.lntts, self.g)

        return self.lntts, self.g
